"""Discrete Surfaces with Elastic metrics.

Lead authors: Emmanuel Hartman, Adele Myers.

References
----------
.. [HSKCB2022] Emmanuel Hartman, Yashil Sukurdeep, Eric Klassen,
    Nicolas Charon, and Martin Bauer.
    "Elastic shape analysis of surfaces with second-order Sobolev metrics:
    a comprehensive numerical framework". arXiv:2204.04238 [cs.CV], 25 Sep 2022
.. [HPBDC2023] Emmanuel Hartman, Emery Pierson, Martin Bauer,
    Mohamed Daoudi, and Nicolas Charon.
    “Basis Restricted Elastic Shape Analysis on the Space of Unregistered Surfaces.”
    arXiv, November 7, 2023. https://doi.org/10.48550/arXiv.2311.04382.
"""

import math

import geomstats.backend as gs
from geomstats._mesh import Surface
from geomstats.geometry.euclidean import Euclidean
from geomstats.geometry.fiber_bundle import AlignerAlgorithm, FiberBundle
from geomstats.geometry.manifold import Manifold, register_quotient
from geomstats.geometry.matrices import Matrices
from geomstats.geometry.nfold_manifold import NFoldMetric
from geomstats.geometry.quotient_metric import QuotientMetric
from geomstats.geometry.riemannian_metric import RiemannianMetric
from geomstats.numerics.geodesic import ExpSolver, PathBasedLogSolver, PathStraightening
from geomstats.numerics.optimization import ScipyMinimize
from geomstats.numerics.path import (
    UniformlySampledDiscretePath,
    UniformlySampledPathEnergy,
)
from geomstats.vectorization import get_batch_shape


class DiscreteSurfaces(Manifold):
    r"""Space of parameterized discrete surfaces.

    Each surface is sampled with fixed `n_vertices` vertices and `n_faces`
    faces in :math:`\mathbb{R}^3`.

    Each individual surface is represented by a 2d-array of shape
    `[n_vertices, 3]`. This space corresponds to the space of immersions
    defined below, i.e. the
    space of smooth functions from a template to manifold :math:`M`
    into :math:`\mathbb{R}^3`, with non-vanishing Jacobian.

    .. math::
        Imm(M,\mathbb{R}^3)=\{ f \in C^{\infty}(M, \mathbb{R}^3)
        \|Df(x)\|\neq 0 \forall x \in M \}.

    Parameters
    ----------
    faces : integer array-like, shape=[n_faces, 3]
        Triangulation of the surface.
        Each face is given by 3 indices that indicate its vertices.
    """

    def __init__(
        self,
        faces,
        equip=True,
    ):
        ambient_dim = 3
        self.ambient_manifold = Euclidean(dim=ambient_dim)
        self.faces = faces
        self.n_faces = len(faces)
        self.n_vertices = int(gs.amax(self.faces) + 1)
        self.shape = (self.n_vertices, ambient_dim)
        super().__init__(
            dim=self.n_vertices * ambient_dim,
            shape=(self.n_vertices, 3),
            equip=equip,
        )

    def new(self, equip=True):
        """Create manifold with same parameters.

        Parameters
        ----------
        equip : bool
            If True, equip space with default metric.

        Returns
        -------
        manifold : DiscreteSurfaces
        """
        return DiscreteSurfaces(faces=self.faces, equip=equip)

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return ElasticMetric

    def belongs(self, point, atol=gs.atol):
        """Evaluate whether a point belongs to the manifold.

        Checks that vertices are inputed in proper form and are
        consistent with the mesh structure.

        Parameters
        ----------
        point : array-like, shape=[..., n_vertices, 3]
            Surface, as the 3D coordinates of the vertices of its triangulation.
        atol : float
            Absolute tolerance.
            Optional, default: backend atol.

        Returns
        -------
        belongs : array-like, shape=[...,]
            Boolean evaluating if point belongs to the space of discrete
            surfaces.
        """
        belongs = self.shape == point.shape[-self.point_ndim :]
        shape = point.shape[: -self.point_ndim]
        if belongs:
            return gs.ones(shape, dtype=bool)
        return gs.zeros(shape, dtype=bool)

    def is_tangent(self, vector, base_point, atol=gs.atol):
        """Check whether the vector is tangent at base_point.

        Tangent vectors are identified with points of the vector space so
        this checks the shape of the input vector.

        Parameters
        ----------
        vector : array-like, shape=[..., n_vertices, 3]
            Vector, i.e. a 3D vector field on the surface.
        base_point : array-like, shape=[..., n_vertices, 3]
            Surface, as the 3D coordinates of the vertices of its triangulation.
        atol : float
            Absolute tolerance.
            Optional, default: backend atol.

        Returns
        -------
        is_tangent : array-like, shape=[...,]
            Boolean denoting if vector is a tangent vector at the base point.
        """
        belongs = self.belongs(vector, atol)
        if base_point is not None and base_point.ndim > vector.ndim:
            return gs.broadcast_to(belongs, base_point.shape[: -self.point_ndim])
        return belongs

    def to_tangent(self, vector, base_point):
        """Project a vector to a tangent space of the manifold.

        Parameters
        ----------
        vector : array-like, shape=[..., n_vertices, 3]
            Vector, i.e. a 3D vector field on the surface.
        base_point : array-like, shape=[..., n_vertices, 3]
            Surface, as the 3D coordinates of the vertices of its triangulation.

        Returns
        -------
        tangent_vec : array-like, shape=[..., *point_shape]
            Tangent vector at base point.
        """
        return gs.copy(vector)

    def projection(self, point):
        """Project a point to the manifold.

        Parameters
        ----------
        point : array-like, shape=[..., n_vertices, 3]
            Surface, as the 3D coordinates of the vertices of its triangulation..

        Returns
        -------
        _ : array-like, shape=[..., n_vertices, 3]
            Point.
        """
        return gs.copy(point)

    def random_point(self, n_samples=1):
        """Sample discrete surfaces.

        This sample random discrete surfaces with the correct number of vertices.

        Parameters
        ----------
        n_samples : int
            Number of surfaces to sample.
            Optional, Default=1

        Returns
        -------
        vertices : array-like, shape=[n_samples, n_vertices, 3]
            Vertices for a batch of points in the space of discrete surfaces.
        """
        vertices = self.ambient_manifold.random_point(n_samples * self.n_vertices)
        vertices = gs.reshape(vertices, (n_samples, self.n_vertices, 3))
        return vertices[0] if n_samples == 1 else vertices

    def _vertices(self, point):
        """Extract 3D vertices coordinates corresponding to each face.

        Parameters
        ----------
        point : array-like, shape=[..., n_vertices, 3]
            Surface, as the 3D coordinates of the vertices of its triangulation.

        Returns
        -------
        vertices : tuple of vertex_0, vertex_1, vertex_2 where:
            vertex_i : array-like, shape=[..., n_faces, 3]
                3D coordinates of the ith vertex of that face.
        """
        batch_slc = tuple([slice(None)] * len(point.shape[:-2]))
        face_coordinates = point[batch_slc + (self.faces,)]
        return (
            face_coordinates[batch_slc + (slice(None), 0)],
            face_coordinates[batch_slc + (slice(None), 1)],
            face_coordinates[batch_slc + (slice(None), 2)],
        )

    def _triangle_areas(self, point):
        """Compute triangle areas for each face of the surface.

        Heron's formula gives the triangle's area in terms of its sides a b c:,
        As the square root of the product s(s - a)(s - b)(s - c),
        where s is the semiperimeter of the triangle s = (a + b + c)/2.

        Parameters
        ----------
        point : array-like, shape=[..., n_vertices, 3]
             Surface, as the 3D coordinates of the vertices of its triangulation.

        Returns
        -------
        _ : array-like, shape=[..., n_faces, 1]
            Triangle area of each face.
        """
        vertex_0, vertex_1, vertex_2 = self._vertices(point)
        len_edge_12 = gs.linalg.norm((vertex_1 - vertex_2), axis=-1)
        len_edge_02 = gs.linalg.norm((vertex_0 - vertex_2), axis=-1)
        len_edge_01 = gs.linalg.norm((vertex_0 - vertex_1), axis=-1)
        half_perimeter = 0.5 * (len_edge_12 + len_edge_02 + len_edge_01)
        return gs.sqrt(
            (
                half_perimeter
                * (half_perimeter - len_edge_12)
                * (half_perimeter - len_edge_02)
                * (half_perimeter - len_edge_01)
            ).clip(min=1e-6)
        )

    def vertex_areas(self, point):
        """Compute vertex areas for a triangulated surface.

        Vertex area is the area of all of the triangles who are in contact (incident)
        with a specific vertex, according to the formula:
        vertex_areas = 2 * sum_incident_areas / 3.0

        Parameters
        ----------
        point : array-like, shape=[..., n_vertices, 3]
             Surface, as the 3D coordinates of the vertices of its triangulation.

        Returns
        -------
        vertex_areas :  array-like, shape=[..., n_vertices]
            Vertex area for each vertex.
        """
        batch_shape = point.shape[:-2]
        n_vertices = point.shape[-2]
        n_faces = self.faces.shape[0]

        area = self._triangle_areas(point)

        id_vertices = gs.broadcast_to(
            gs.reshape(self.faces, (-1,)), batch_shape + (math.prod(self.faces.shape),)
        )
        val = gs.reshape(
            gs.broadcast_to(gs.expand_dims(area, axis=-2), batch_shape + (3, n_faces)),
            batch_shape + (-1,),
        )
        incident_areas = gs.zeros(batch_shape + (n_vertices,), dtype=val.dtype)

        incident_areas = gs.scatter_add(
            incident_areas,
            dim=-1,
            index=id_vertices,
            src=val,
        )
        return 2 * incident_areas / 3.0

    def normals(self, point):
        """Compute normals at each face of a triangulated surface.

        Normals are the cross products between edges of each face
        that are incident to its x-coordinate.

        Parameters
        ----------
        point : array-like, shape=[..., n_vertices, 3]
            Surface, as the 3D coordinates of the vertices of its triangulation.

        Returns
        -------
        normals_at_point : array-like, shape=[..., n_faces, 3]
            Normals of each face of the mesh.
        """
        vertex_0, vertex_1, vertex_2 = self._vertices(point)
        return 0.5 * gs.cross(vertex_1 - vertex_0, vertex_2 - vertex_0)

    def surface_one_forms(self, point):
        """Compute the vector valued one-forms.

        The one forms are evaluated at the faces of a triangulated surface.

        A one-form is represented by the two vectors (01) and (02) at each face
        of vertices 0, 1, 2.

        Parameters
        ----------
        point : array-like, shape=[..., n_vertices, 3]
             Surface, as the 3D coordinates of the vertices of its triangulation.

        Returns
        -------
        one_forms_bp : array-like, shape=[..., n_faces, 2, 3]
            One form evaluated at each face of the triangulated surface.
        """
        vertex_0, vertex_1, vertex_2 = self._vertices(point)
        return gs.stack([vertex_1 - vertex_0, vertex_2 - vertex_0], axis=-2)

    def face_areas(self, point):
        """Compute the areas for each face of a triangulated surface.

        The corresponds to the volume area for the surface metric, that is
        the volume area of the pullback metric of the immersion defining the
        surface metric.

        Parameters
        ----------
        point : array-like, shape=[..., n_vertices, 3]
            Surface, as the 3D coordinates of the vertices of its triangulation.

        Returns
        -------
        _ : array-like, shape=[..., n_faces,]
            Area computed at each face of the triangulated surface.
        """
        surface_metrics_bp = self.surface_metric_matrices(point)
        return gs.sqrt(gs.linalg.det(surface_metrics_bp))

    @staticmethod
    def surface_metric_matrices_from_one_forms(one_forms):
        """Compute the surface metric matrices directly from the one_forms.

        This function is useful for efficiency purposes.

        Parameters
        ----------
        one_forms : array-like, shape=[..., n_faces, 2, 3]
            One form evaluated at each face of the triangulated surface.

        Returns
        -------
        metric_mats : array-like, shape=[..., n_faces, 2, 2]
            Surface metric matrices evaluated at each face of
            the triangulated surface.
        """
        return gs.matmul(one_forms, Matrices.transpose(one_forms))

    def surface_metric_matrices(self, point):
        """Compute the surface metric matrices.

        The matrices are evaluated at the faces of a triangulated surface.

        The surface metric is the pullback metric of the immersion q
        defining the surface, i.e. of
        the map q: M -> R3, where M is the parameterization manifold.

        Parameters
        ----------
        point : array like, shape=[..., n_vertices, 3]
            Surface, as the 3D coordinates of the vertices of its triangulation.

        Returns
        -------
        metric_mats : array-like, shape=[..., n_faces, 2, 2]
            Surface metric matrices evaluated at each face of
            the triangulated surface.
        """
        one_forms = self.surface_one_forms(point)
        return self.surface_metric_matrices_from_one_forms(one_forms)

    def laplacian(self, point):
        r"""Compute the mesh Laplacian operator of a triangulated surface.

        Denoting q the surface, i.e. the point in the DiscreteSurfaces manifold,
        the laplacian at :math:`q` is defined as the operator:
        :math:`\Delta_q = - Tr(g_q^{-1} \nabla^2)`
        where :math:`g_q` is the surface metric matrix of :math:`q`.

        The area of the triangles is computed using Heron's formula.

        Parameters
        ----------
        point :  array-like, shape=[..., n_vertices, 3]
            Surface, as the 3D coordinates of the vertices of its triangulation.

        Returns
        -------
        _laplacian : callable
            Function that evaluates the mesh Laplacian operator at a
            tangent vector field to the surface.
        """
        n_vertices, n_faces = point.shape[-2], self.faces.shape[0]
        vertex_0, vertex_1, vertex_2 = self._vertices(point)
        len_edge_12 = gs.linalg.norm((vertex_1 - vertex_2), axis=-1)
        len_edge_02 = gs.linalg.norm((vertex_0 - vertex_2), axis=-1)
        len_edge_01 = gs.linalg.norm((vertex_0 - vertex_1), axis=-1)

        half_perimeter = 0.5 * (len_edge_12 + len_edge_02 + len_edge_01)
        area = gs.sqrt(
            (
                half_perimeter
                * (half_perimeter - len_edge_12)
                * (half_perimeter - len_edge_02)
                * (half_perimeter - len_edge_01)
            ).clip(min=1e-6)
        )
        sq_len_edge_12, sq_len_edge_02, sq_len_edge_01 = (
            len_edge_12 * len_edge_12,
            len_edge_02 * len_edge_02,
            len_edge_01 * len_edge_01,
        )
        cot_12 = (sq_len_edge_02 + sq_len_edge_01 - sq_len_edge_12) / area
        cot_02 = (sq_len_edge_12 + sq_len_edge_01 - sq_len_edge_02) / area
        cot_01 = (sq_len_edge_12 + sq_len_edge_02 - sq_len_edge_01) / area
        cot = gs.stack([cot_12, cot_02, cot_01], axis=-1)
        cot /= 2.0
        id_vertices_120 = self.faces[:, [1, 2, 0]]
        id_vertices_201 = self.faces[:, [2, 0, 1]]
        id_vertices = gs.reshape(
            gs.stack([id_vertices_120, id_vertices_201], axis=0), (2, n_faces * 3)
        )

        cot_flatten = gs.expand_dims(gs.reshape(cot, point.shape[:-2] + (-1,)), axis=-1)

        def _laplacian(tangent_vec):
            r"""Evaluate the mesh Laplacian operator.

            The operator is evaluated at a tangent vector at point to the
            manifold of DiscreteSurfaces. In other words, the operator is
            evaluated at a vector field defined on the surface point.

            .. math::

                \left(\Delta_q h\right)_{v_i}=\frac{1}{2} \sum_{\substack{j \mid(i, j)
                \in E \\ \text { or }(j, i) \in E}}\left(\cot \left(\alpha_{i j}
                \right)+\cot \left(\beta_{i j}\right)\right)\left(h_i-h_j\right)

            Parameters
            ----------
            tangent_vec : array-like, shape=[..., n_vertices, 3]
                Tangent vector to the manifold at the base point that is the
                triangulated surface. This tangent vector is a vector field
                on the triangulated surface.

            Returns
            -------
            laplacian_at_tangent_vec: array-like, shape=[..., n_vertices, 3]
                Mesh Laplacian operator of the triangulated surface applied
                to one its tangent vector tangent_vec.
            """
            batch_shape = get_batch_shape(2, point, tangent_vec)
            batch_slc = tuple([slice(None)] * len(batch_shape))

            tangent_vec_diff = (
                tangent_vec[batch_slc + (id_vertices[0],)]
                - tangent_vec[batch_slc + (id_vertices[1],)]
            )

            values = gs.einsum(
                "...bd,...bd->...bd",
                gs.broadcast_to(cot_flatten, batch_shape + (n_faces * 3, 3)),
                tangent_vec_diff,
            )

            laplacian_at_tangent_vec = gs.zeros(
                batch_shape + (n_vertices, 3), dtype=values.dtype
            )

            id_vertices_201 = id_vertices[1, :]
            id_vertices_201 = gs.broadcast_to(
                id_vertices_201, batch_shape + id_vertices_201.shape
            )

            for i_dim in range(3):
                laplacian_at_tangent_vec[batch_slc + (slice(None), i_dim)] = (
                    gs.scatter_add(
                        input=laplacian_at_tangent_vec[
                            batch_slc + (slice(None), i_dim)
                        ],
                        dim=-1,
                        index=id_vertices_201,
                        src=values[batch_slc + (slice(None), i_dim)],
                    )
                )
            return laplacian_at_tangent_vec

        return _laplacian


class ElasticMetric(RiemannianMetric):
    """Elastic metric defined by a family of second order Sobolev metrics.

    Each individual discrete surface is represented by a 2D-array of shape
    `[n_vertices, 3]`. See [HSKCB2022]_ and [HPBDC2023]_ (appendix) for details.

    The parameters a0, a1, b1, c1, d1, a2 (detailed below) are non-negative weighting
    coefficients for the different terms in the metric.

    Parameters
    ----------
    space : DiscreteSurfaces
        Instantiated DiscreteSurfaces manifold.
    a0 : float
        First order parameter.
        Default: 1.
    a1 : float
        Stretching parameter.
        Default: 1.
    b1 : float
        Shearing parameter.
        Default: 1.
    c1 : float
        Bending parameter.
        Default: 1.
    d1 : float
        Additonal first order parameter.
        Default: 1.
    a2 : float
        Second order parameter.
        Default: 1.

    References
    ----------
    .. [HSKCB2022] Emmanuel Hartman, Yashil Sukurdeep, Eric Klassen,
        Nicolas Charon, and Martin Bauer.
        "Elastic shape analysis of surfaces with second-order Sobolev metrics:
        a comprehensive numerical framework". arXiv:2204.04238 [cs.CV], 25 Sep 2022
    .. [HPBDC2023] Emmanuel Hartman, Emery Pierson, Martin Bauer,
        Mohamed Daoudi, and Nicolas Charon.
        “Basis Restricted Elastic Shape Analysis on the Space of Unregistered Surfaces.”
        arXiv, November 7, 2023. https://doi.org/10.48550/arXiv.2311.04382.
    """

    def __init__(self, space, a0=1.0, a1=1.0, b1=1.0, c1=1.0, d1=1.0, a2=1.0):
        super().__init__(space=space)
        self.a0 = a0
        self.a1 = a1
        self.b1 = b1
        self.c1 = c1
        self.d1 = d1
        self.a2 = a2

        if gs.has_autodiff():
            self.exp_solver = DiscreteSurfacesExpSolver(space, n_steps=10)

            optimizer = ScipyMinimize(
                method="L-BFGS-B",
                autodiff_jac=True,
                options={"disp": False, "ftol": 0.001},
            )
            self.log_solver = PathStraightening(space, n_nodes=10, optimizer=optimizer)

    def _inner_product_a0(self, tangent_vec_a, tangent_vec_b, vertex_areas_bp):
        r"""Compute term of order 0 within the inner-product.

        Denote h and k the tangent vectors a and b respectively.
        Denote q the base point.

        This method computes :math:`G_{a_0} = a_0 <h, k>`, i.e.

        .. math::

            G_{a_0}(h, h) = \sum_{i=1}^N\left\|h_i\right\|^2 \operatorname{vol}_{x_i}

        Parameters
        ----------
        tangent_vec_a : array-like, shape=[..., n_vertices, 3]
            Tangent vector at base point.
        tangent_vec_b : array-like, shape=[..., n_vertices, 3]
            Tangent vector at base point.
        vertex_areas : array-like, shape=[..., n_vertices, 1]
            Vertex areas for each vertex of the base_point.

        Returns
        -------
        inner_prod_a0 : array-like, shape=[...,]
            Term of order 0, and coefficient a0, of the inner-product.
        """
        return self.a0 * gs.sum(
            gs.dot(tangent_vec_a, tangent_vec_b) * vertex_areas_bp,
            axis=-1,
        )

    def _inner_product_a1(self, ginvdga, ginvdgb, areas_bp):
        r"""Compute a1 term of order 1 within the inner-product.

        Denote h and k the tangent vectors a and b respectively.
        Denote q the base point.

        This method computes :math:`G_{a_1} = a_1.g_q^{-1} <dh_m, dk_m>`, i.e.

        .. math::

            G_{a_1}(h, h) = \sum_{f \in F} \operatorname{tr}\left(g_f^{-1}
            \delta g_f g_f^{-1} \delta g_f\right) \operatorname{vol}_f

        Parameters
        ----------
        ginvdga : array-like, shape=[..., n_faces, 2, 2]
            Product of the inverse of the surface metric matrices
            with their differential at a.
        ginvdgb : array-like, shape=[..., n_faces, 2, 2]
            Product of the inverse of the surface metric matrices
            with their differential at b.
        areas_bp : array-like, shape=[..., n_faces,]
            Areas of the faces of the surface given by the base point.

        Returns
        -------
        inner_prod_a1 : array-like, shape=[...,]
            Term of order 1, and coefficient a1, of the inner-product.
        """
        return self.a1 * gs.sum(
            gs.trace(gs.matmul(ginvdga, ginvdgb)) * areas_bp,
            axis=-1,
        )

    def _inner_product_b1(self, ginvdga, ginvdgb, areas_bp):
        r"""Compute b1 term of order 1 within the inner-product.

        Denote h and k the tangent vectors a and b respectively.
        Denote q the base point.

        This method computes :math:`G_{b_1} = b_1.g_q^{-1} <dh_+, dk_+>`, i.e.

        .. math::

            G_{b_1}(h, h) = \sum_{f \in F} \operatorname{tr}
            \left(g_f^{-1} \delta g_f\right)^2 \operatorname{vol}_f

        Parameters
        ----------
        ginvdga : array-like, shape=[..., n_faces, 2, 2]
            Product of the inverse of the surface metric matrices
            with their differential at a.
        ginvdgb : array-like, shape=[..., n_faces, 2, 2]
            Product of the inverse of the surface metric matrices
            with their differential at b.
        areas_bp : array-like, shape=[..., n_faces,]
            Areas of the faces of the surface given by the base point.

        Returns
        -------
        inner_prod_b1 : array-like, shape=[...,]
            Term of order 1, and coefficient b1, of the inner-product.
        """
        return self.b1 * gs.sum(
            gs.trace(ginvdga) * gs.trace(ginvdgb) * areas_bp,
            axis=-1,
        )

    def _inner_product_c1(self, point_a, point_b, normals_bp, areas_bp):
        r"""Compute c1 term of order 1 within the inner-product.

        Denote h and k the tangent vectors a and b respectively.
        Denote q the base point.

        This method computes :math:`G_{c_1} = c_1.g_q^{-1} <dh_\perp, dk_\perp>`,
        i.e.

        .. math::

            G_{c_1}(h, h) = \sum_{f \in F}\left\langle\delta n_f,
            \delta n_f\right\rangle \mathrm{vol}_f

        Parameters
        ----------
        point_a : array-like, shape=[..., n_vertices, 3]
            Point a corresponding to tangent vec a.
        point_b : array-like, shape=[..., n_vertices, 3]
            Point b corresponding to tangent vec b.
        normals_bp : array-like, shape=[..., n_faces, 3]
            Normals of each face of the surface given by the base point.
        areas_bp : array-like, shape=[..., n_faces,]
            Areas of the faces of the surface given by the base point.

        Returns
        -------
        inner_prod_c1 : array-like, shape=[...,]
            Term of order 1, and coefficient c1, of the inner-product.
        """
        dna = self._space.normals(point_a) - normals_bp
        dnb = self._space.normals(point_b) - normals_bp

        return self.c1 * gs.sum(gs.dot(dna, dnb) * areas_bp, axis=-1)

    def _inner_product_d1(
        self, one_forms_a, one_forms_b, one_forms_bp, areas_bp, ginv_bp
    ):
        r"""Compute d1 term of order 1 within the inner-product.

        Denote h and k the tangent vectors a and b respectively.
        Denote q the base point.

        This method computes :math:`G_{d_1} = d_1.g_q^{-1} <dh_0, dk_0>`, i.e.

        .. math::

            G_{d_1}(h, h) = \sum_{f \in F} \operatorname{tr}\left(g_f^{-1}
            \xi_f g_f^{-1} \xi_f^T\right) \operatorname{vol}_f

        where :math:`\xi_f=d q_f^T d h_f-d h_f^T d q_f`.

        Parameters
        ----------
        one_forms_a : array-like, shape=[..., n_faces, 2, 3]
            One forms at point a corresponding to tangent vec a.
        one_forms_b : array-like, shape=[..., n_faces, 2, 3]
            One forms at point b corresponding to tangent vec b.
        one_forms_bp : array-like, shape=[..., 2, 3]
            One forms at base point.
        areas_bp : array-like, shape=[..., n_faces,]
            Areas of the faces of the surface given by the base point.
        ginv_bp : array-like, shape=[..., n_faces, 2, 2]
            Inverses of the surface metric matrices at each face.

        Returns
        -------
        inner_prod_d1 : array-like, shape=[...,]
            Term of order 1, and coefficient d1, of the inner-product.
        """
        one_forms_bp_t = Matrices.transpose(one_forms_bp)

        aux = gs.matmul(one_forms_bp_t, ginv_bp)

        xa = one_forms_a - one_forms_bp
        xa_0 = gs.matmul(
            aux,
            gs.matmul(xa, one_forms_bp_t)
            - gs.matmul(one_forms_bp, Matrices.transpose(xa)),
        )

        xb = one_forms_b - one_forms_bp
        xb_0 = gs.matmul(
            aux,
            gs.matmul(xb, one_forms_bp_t)
            - gs.matmul(one_forms_bp, Matrices.transpose(xb)),
        )

        return self.d1 * gs.sum(
            gs.trace(
                Matrices.mul(
                    xa_0,
                    ginv_bp,
                    Matrices.transpose(xb_0),
                ),
            )
            * areas_bp,
            axis=-1,
        )

    def _inner_product_a2(
        self, tangent_vec_a, tangent_vec_b, base_point, vertex_areas_bp
    ):
        r"""Compute term of order 2 within the inner-product.

        Denote h and k the tangent vectors a and b respectively.
        Denote q the base point.

        This method computes :math:`G_{a_2} = a_2 <\Delta_q h, \Delta_q k>`, i.e.

        .. math::

            G_{a_2}(h, h) = \sum_{i=1}^N\left\|\left(\Delta_q h\right)_{v_i}
            \right\|^2 \operatorname{vol}_{x_i}


        Parameters
        ----------
        tangent_vec_a : array-like, shape=[..., n_vertices, 3]
            Tangent vector at base point.
        tangent_vec_b : array-like, shape=[..., n_vertices, 3]
            Tangent vector at base point.
        base_point : array-like, shape=[..., n_vertices, 3]
            Base point, a surface i.e. the 3D coordinates of its vertices.
        vertex_areas_bp : array-like, shape=[..., n_vertices, 1]
            Vertex areas for each vertex of the base_point.

        Returns
        -------
        inner_prod_a2 : array-like, shape=[...,]
            Term of order 2, and coefficient a2, of the inner-product.
        """
        laplacian_at_base_point = self._space.laplacian(base_point)
        return self.a2 * gs.sum(
            gs.dot(
                laplacian_at_base_point(tangent_vec_a),
                laplacian_at_base_point(tangent_vec_b),
            )
            / vertex_areas_bp,
            axis=-1,
        )

    def inner_product(self, tangent_vec_a, tangent_vec_b, base_point):
        r"""Compute inner product between two tangent vectors at a base point.

        The inner-product has 6 terms, where each term corresponds to
        one of the 6 hyperparameters a0, a1, b1, c1, d1, a2.

        We denote h and k the tangent vectors a and b respectively.
        We denote q the base point, i.e. the surface.

        The six terms of the inner-product are given by:

        .. math::

            \int_M (G_{a_0} + G_{a_1} + G_{b_1} + G_{c_1} + G_{d_1} + G_{a_2})vol_q

        where:

        - :math:`G_{a_0} = a_0 <h, k>`
        - :math:`G_{a_1} = a_1.g_q^{-1} <dh_m, dk_m>`
        - :math:`G_{b_1} = b_1.g_q^{-1} <dh_+, dk_+>`
        - :math:`G_{c_1} = c_1.g_q^{-1} <dh_\perp, dk_\perp>`
        - :math:`G_{d_1} = d_1.g_q^{-1} <dh_0, dk_0>`
        - :math:`G_{a_2} = a_2 <\Delta_q h, \Delta_q k>`

        Parameters
        ----------
        tangent_vec_a : array-like, shape=[..., n_vertices, 3]
            Tangent vector at base point.
        tangent_vec_b : array-like, shape=[..., n_vertices, 3]
            Tangent vector at base point.
        base_point : array-like, shape=[..., n_vertices, 3]
            Surface, as the 3D coordinates of the vertices of its triangulation.

        Returns
        -------
        inner_prod : array-like, shape=[...]
            Inner product.
        """
        inner_prod_a0 = 0.0
        inner_prod_a1 = 0.0
        inner_prod_a2 = 0.0
        inner_prod_b1 = 0.0
        inner_prod_c1 = 0.0
        inner_prod_d1 = 0.0

        if self.a0 > 0 or self.a2 > 0:
            vertex_areas_bp = self._space.vertex_areas(base_point)
            if self.a0 > 0:
                inner_prod_a0 = self._inner_product_a0(
                    tangent_vec_a, tangent_vec_b, vertex_areas_bp=vertex_areas_bp
                )
            if self.a2 > 0:
                inner_prod_a2 = self._inner_product_a2(
                    tangent_vec_a,
                    tangent_vec_b,
                    base_point=base_point,
                    vertex_areas_bp=vertex_areas_bp,
                )
        if self.a1 > 0 or self.b1 > 0 or self.c1 > 0 or self.b1 > 0:
            one_forms_bp = self._space.surface_one_forms(base_point)
            surface_metrics_bp = self._space.surface_metric_matrices_from_one_forms(
                one_forms_bp
            )
            areas_bp = gs.sqrt(gs.linalg.det(surface_metrics_bp))

            point_a = base_point + tangent_vec_a
            point_b = base_point + tangent_vec_b

            if self.c1 > 0:
                normals_bp = self._space.normals(base_point)
                inner_prod_c1 = self._inner_product_c1(
                    point_a, point_b, normals_bp, areas_bp
                )
            if self.d1 > 0 or self.b1 > 0 or self.a1 > 0:
                ginv_bp = gs.linalg.inv(surface_metrics_bp)
                one_forms_a = self._space.surface_one_forms(point_a)
                one_forms_b = self._space.surface_one_forms(point_b)
                if self.d1 > 0:
                    inner_prod_d1 = self._inner_product_d1(
                        one_forms_a,
                        one_forms_b,
                        one_forms_bp,
                        areas_bp,
                        ginv_bp,
                    )

                if self.b1 > 0 or self.a1 > 0:
                    surface_metrics_a = (
                        self._space.surface_metric_matrices_from_one_forms(one_forms_a)
                    )
                    surface_metrics_b = (
                        self._space.surface_metric_matrices_from_one_forms(one_forms_b)
                    )
                    dga = surface_metrics_a - surface_metrics_bp
                    dgb = surface_metrics_b - surface_metrics_bp
                    ginvdga = gs.matmul(ginv_bp, dga)
                    ginvdgb = gs.matmul(ginv_bp, dgb)
                    if self.a1 > 0:
                        inner_prod_a1 = self._inner_product_a1(
                            ginvdga, ginvdgb, areas_bp
                        )
                    if self.b1 > 0:
                        inner_prod_b1 = self._inner_product_b1(
                            ginvdga, ginvdgb, areas_bp
                        )

        return (
            inner_prod_a0
            + inner_prod_a1
            + inner_prod_a2
            + inner_prod_b1
            + inner_prod_c1
            + inner_prod_d1
        )


class DiscreteSurfacesExpSolver(ExpSolver):
    """Class to solve the initial value problem (IVP) for exp.

    Implements methods from discrete geodesic calculus method.
    """

    def __init__(self, space, n_steps=10, optimizer=None):
        super().__init__(solves_ivp=True)
        self._space = space
        if optimizer is None:
            optimizer = ScipyMinimize(
                method="L-BFGS-B",
                autodiff_jac=True,
                options={"disp": False, "ftol": 0.00001},
            )

        self.n_steps = n_steps
        self.optimizer = optimizer

    def _objective(self, current_point, next_point):
        """Return objective function to compute the next point on the geodesic.

        Parameters
        ----------
        current_point : array-like, shape=[n_vertices, 3]
            Current point on the geodesic.
        next_point : array-like, shape=[n_vertices, 3]
            Next point on the geodesic.

        Returns
        -------
        energy_objective : callable
            Computes energy wrt next next point.
        """
        zeros = gs.zeros_like(current_point)

        def energy_objective(flat_next_next_point):
            """Compute the energy objective to minimize.

            Parameters
            ----------
            next_next_point : array-like, shape=[n_vertices*3]
                Next next point on the geodesic.

            Returns
            -------
            energy_tot : array-like, shape=[,]
                Energy objective to minimize.
            """
            next_next_point = gs.reshape(flat_next_next_point, self._space.shape)
            current_to_next = next_point - current_point
            next_to_next_next = next_next_point - next_point

            def _inner_product_with_current_to_next(tangent_vec):
                """Compute inner-product with tangent vector `current_to_next`.

                The tangent vector `current_to_next` is the vector going from the
                current point, i.e. discrete surface, to the next point on the
                geodesic that is being computed.
                """
                return self._space.metric.inner_product(
                    current_to_next, tangent_vec, current_point
                )

            def _inner_product_with_next_to_next_next(tangent_vec):
                """Compute inner-product with tangent vector `next_to_next_next`.

                The tangent vector `next_to_next_next` is the vector going from the
                next point, i.e. discrete surface, to the next next point on the
                geodesic that is being computed.
                """
                return self._space.metric.inner_product(
                    next_to_next_next, tangent_vec, next_point
                )

            def _norm(base_point):
                """Compute norm of `next_to_next_next` at the base_point.

                The tangent vector `next_to_next_next` is the vector going from the
                next point, i.e. discrete surface, to the next next point on the
                geodesic that is being computed.
                """
                return self._space.metric.squared_norm(next_to_next_next, base_point)

            _, energy_1 = gs.autodiff.value_and_grad(
                _inner_product_with_current_to_next,
                point_ndims=2,
            )(zeros)
            _, energy_2 = gs.autodiff.value_and_grad(
                _inner_product_with_next_to_next_next,
                point_ndims=2,
            )(zeros)
            _, energy_3 = gs.autodiff.value_and_grad(_norm, point_ndims=2)(next_point)

            energy_tot = 2 * energy_1 - 2 * energy_2 + energy_3
            return gs.sum(energy_tot**2)

        return energy_objective

    def _stepforward(self, current_point, next_point):
        """Compute the next point on the geodesic.

        Parameters
        ----------
        current_point : array-like, shape=[n_vertices, 3]
            Current point on the geodesic.
        next_point : array-like, shape=[n_vertices, 3]
            Next point on the geodesic.

        Returns
        -------
        next_next_point : array-like, shape=[n_vertices, 3]
            Next next point on the geodesic.
        """
        flat_initial_next_next_point = gs.flatten(
            (2 * (next_point - current_point) + current_point)
        )

        energy_objective = self._objective(current_point, next_point)

        sol = self.optimizer.minimize(
            energy_objective,
            flat_initial_next_next_point,
        )

        return gs.reshape(sol.x, self._space.shape)

    def _discrete_geodesic_ivp_single(self, tangent_vec, base_point):
        """Solve initial value problem (IVP).

        Given an initial tangent vector and an initial point,
        solve the geodesic equation.

        Parameters
        ----------
        tangent_vec : array-like, shape=[n_vertices, 3]
            Initial tangent vector.
        base_point : array-like, shape=[n_vertices, 3]
            Initial point, i.e. initial discrete surface.

        Returns
        -------
        geod : array-like, shape=[n_steps, n_vertices, 3]
            Discretized geodesic uniformly sampled.
        """
        next_point = base_point + tangent_vec / (self.n_steps - 1)
        geod = [base_point, next_point]
        for _ in range(2, self.n_steps):
            next_next_point = self._stepforward(geod[-2], geod[-1])
            geod.append(next_next_point)

        return gs.stack(geod, axis=0)

    def discrete_geodesic_ivp(self, tangent_vec, base_point):
        """Solve initial value problem (IVP).

        Given an initial tangent vector and an initial point,
        solve the geodesic equation.

        Parameters
        ----------
        tangent_vec : array-like, shape=[n_vertices, 3]
            Initial tangent vector.
        base_point : array-like, shape=[n_vertices, 3]
            Initial point, i.e. initial discrete surface.

        Returns
        -------
        geod : array-like, shape=[n_steps, n_vertices, 3]
            Discretized geodesic uniformly sampled.
        """
        if tangent_vec.ndim != base_point.ndim:
            tangent_vec, base_point = gs.broadcast_arrays(tangent_vec, base_point)

        is_batch = base_point.ndim > self._space.point_ndim
        if not is_batch:
            return self._discrete_geodesic_ivp_single(tangent_vec, base_point)

        return gs.stack(
            [
                self._discrete_geodesic_ivp_single(tangent_vec_, base_point_)
                for tangent_vec_, base_point_ in zip(tangent_vec, base_point)
            ]
        )

    def exp(self, tangent_vec, base_point):
        """Compute exponential map associated to the Riemmannian metric.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., n_vertices, 3]
            Tangent vector at the base point.
        base_point : array-like, shape=[..., n_vertices, 3]
            Point on the manifold, i.e.

        Returns
        -------
        point : array-like, shape=[..., n_vertices, 3]
            Point on the manifold.
        """
        discr_geod_path = self.discrete_geodesic_ivp(tangent_vec, base_point)
        return discr_geod_path[..., -1, :, :]

    def geodesic_ivp(self, tangent_vec, base_point):
        """Geodesic curve for initial value problem.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., n_vertices, 3]
            Initial tangent vector.
        base_point : array-like, shape=[..., n_vertices, 3]
            Initial point, i.e. initial discrete surface.

        Returns
        -------
        path : callable
            Time parametrized geodesic curve. `f(t)`.
        """
        discr_geod_path = self.discrete_geodesic_ivp(tangent_vec, base_point)
        return UniformlySampledDiscretePath(
            discr_geod_path, point_ndim=self._space.point_ndim
        )


class L2SurfacesMetric(NFoldMetric):
    """L2 metric on the space of discrete surfaces.

    Notes
    -----
    Adds ``base_manifold`` (manifold that is repeated ``n_copies``
    times) and ``n_copies`` to ``space`` as attribute if it does
    not have them as their expected by ``NFoldMetric``.
    """

    def __init__(self, space):
        if not hasattr(space, "base_manifold"):
            space.base_manifold = Euclidean(dim=3)
            space.n_copies = space.n_vertices

        super().__init__(space=space)

    def inner_product(self, tangent_vec_a, tangent_vec_b, base_point):
        """Compute L2 inner product between two tangent vectors.

        The inner product is the integral of the ambient space inner product,
        approximated by a left Riemann sum.

        Parameters
        ----------
        tangent_vec_a : array-like, shape=[..., n_vertices, 3]
            Tangent vector at base point.
        tangent_vec_b : array-like, shape=[..., n_vertices, 3]
            Tangent vector at base point.
        base_point : array-like, shape=[..., n_vertices, 3]
            Surface, as the 3D coordinates of the vertices of its triangulation.

        Return
        ------
        inner_prod : array_like, shape=[...]
            Inner product.
        """
        inner_products = self.pointwise_inner_product(
            tangent_vec_a, tangent_vec_b, base_point
        )
        dt = 1 / self._space.n_vertices
        return dt * gs.sum(inner_products, axis=-1)


class RelaxedPathStraightening(PathBasedLogSolver, AlignerAlgorithm):
    """Class to solve the geodesic boundary value problem with path-straightening.

    Parameters
    ----------
    space : Manifold
        Equipped manifold.
    n_nodes : int
        Number of path discretization points.
    lambda_ : float
        Discrepancy loss weight.
    discrepancy_loss : callable
        A generic discrepancy term. Receives point and outputs a callable
        which receives another point and outputs a scalar measuring
        discrepancy between point and another point.
        Scalar sums to energy of a path.
    path_energy : callable
        Method to compute Riemannian path energy.
    optimizer : ScipyMinimize
        An optimizer to solve path energy minimization problem.
    initialization : callable or array-like, shape=[n_nodes - 2, n_vertices, 3]
        A method to get initial guess for optimization or an initial path.

    References
    ----------
    .. [HSKCB2022] "Elastic shape analysis of surfaces with second-order
        Sobolev metrics: a comprehensive numerical framework".
        arXiv:2204.04238 [cs.CV], 25 Sep 2022
    """

    def __init__(
        self,
        total_space,
        n_nodes=3,
        lambda_=1.0,
        discrepancy_loss=None,
        path_energy=None,
        optimizer=None,
        initialization=None,
    ):
        PathBasedLogSolver.__init__(self, space=total_space)
        AlignerAlgorithm.__init__(self, total_space=total_space)

        if discrepancy_loss is None:
            discrepancy_loss = self._default_discrepancy_loss()

        if path_energy is None:
            path_energy = UniformlySampledPathEnergy(total_space)

        if initialization is None:
            initialization = self._default_initialization

        if optimizer is None:
            optimizer = ScipyMinimize(
                method="L-BFGS-B",
                autodiff_jac=True,
                options={"disp": False},
            )

        self.n_nodes = n_nodes
        self.lambda_ = lambda_
        self.discrepancy_loss = discrepancy_loss
        self.path_energy = path_energy
        self.optimizer = optimizer
        self.initialization = initialization

    def _default_discrepancy_loss(self):
        """Discrepancy loss by default.

        Returns
        -------
        loss : callable
            Discrepancy term loss: `f(point) -> g(another_point)`.
            Returns a callable that computes the discrepancy term between
            `point` and `another_point`.
        """
        from geomstats.varifold import (
            BinetKernel,
            GaussianKernel,
            SurfacesKernel,
            VarifoldMetric,
        )

        position_kernel = GaussianKernel(sigma=1.0, init_index=0)
        tangent_kernel = BinetKernel(init_index=position_kernel.new_variable_index())
        kernel = SurfacesKernel(
            position_kernel,
            tangent_kernel,
        )
        varifold_metric = VarifoldMetric(kernel)
        return lambda point: varifold_metric.loss(
            point, target_faces=self._total_space.faces
        )

    def _default_initialization(self, point, base_point):
        """Linear initialization.

        Parameters
        ----------
        point : array-like, shape=[n_vertices, 3]
        base_point : array-like, shape=[n_vertices, 3]

        Returns
        -------
        midpoints : array-like, shape=[n_nodes - 1, n_vertices, 3]
        """
        return gs.broadcast_to(
            base_point, (self.n_nodes - 1,) + self._total_space.shape
        )

    def _discrete_geodesic_bvp_single(self, point, base_point):
        """Solve boundary value problem (BVP).

        Given an initial point and an end point, solve the geodesic equation
        via minimizing the Riemannian path energy.

        Parameters
        ----------
        point : Surface or list[Surface] or array-like, shape=[n_vertices, 3]
        base_point : array-like, shape=[n_vertices, 3]

        Returns
        -------
        discr_geod_path : array-like, shape=[n_times, n_nodes, *point_shape]
            Discrete geodesic.
        """
        if callable(self.initialization):
            init_midpoints = self.initialization(point, base_point)
        else:
            init_midpoints = self.initialization

        base_point = gs.expand_dims(base_point, axis=0)

        discrepancy_loss = self.discrepancy_loss(point)

        def objective(midpoints):
            """Compute path energy of paths going through a midpoint.

            Parameters
            ----------
            midpoint : array-like, shape=[(self.n_nodes-1) * math.prod(n_vertices*3)]
                Midpoints of the path.

            Returns
            -------
            _ : array-like, shape=[...,]
                Energy of the path going through this midpoint.
            """
            midpoints = gs.reshape(
                midpoints, (self.n_nodes - 1,) + self._total_space.shape
            )
            path = gs.concatenate(
                [
                    base_point,
                    midpoints,
                ],
            )
            return self.path_energy(path) + self.lambda_ * discrepancy_loss(path[-1])

        init_midpoints = gs.reshape(init_midpoints, (-1,))
        sol = self.optimizer.minimize(objective, init_midpoints)

        solution_midpoints = gs.reshape(
            gs.array(sol.x), (self.n_nodes - 1,) + self._total_space.shape
        )

        return gs.concatenate(
            [
                base_point,
                solution_midpoints,
            ],
            axis=0,
        )

    def discrete_geodesic_bvp(self, point, base_point):
        """Solve boundary value problem (BVP).

        Given an initial point and an end point, solve the geodesic equation
        via minimizing the Riemannian path energy and a relaxation term.

        Parameters
        ----------
        point : Surface or list[Surface] or array-like, shape=[..., n_vertices, 3]
        base_point : Surface or list[Surface] or array-like, shape=[..., n_vertices, 3]

        Returns
        -------
        discr_geod_path : array-like, shape=[..., n_times, n_nodes, n_vertices, 3]
            Discrete geodesic.
        """
        if not gs.is_array(base_point):
            if _is_iterable(base_point):
                base_point = gs.stack([point_.vertices for point_ in base_point])
            else:
                base_point = base_point.vertices

        if gs.is_array(point):
            if point.ndim > 2:
                point = [
                    Surface(point_, faces=self._total_space.faces) for point_ in point
                ]
            else:
                point = Surface(point, faces=self._total_space.faces)

        if base_point.ndim == 2 and not _is_iterable(point):
            return self._discrete_geodesic_bvp_single(point, base_point)

        if base_point.ndim == 2:
            batch_shape = (len(base_point),)
            base_point = gs.broadcast_to(
                base_point, batch_shape + self._total_space.shape
            )

        elif not _is_iterable(point):
            point = [point] * base_point.shape[-3]

        return gs.stack(
            [
                self._discrete_geodesic_bvp_single(point_, base_point_)
                for base_point_, point_ in zip(base_point, point)
            ]
        )

    def align(self, point, base_point):
        """Align point to base point.

        Parameters
        ----------
        point : Surface or list[Surface] or array-like, shape=[..., n_vertices, 3]
        base_point : array-like, shape=[..., n_vertices, 3]

        Returns
        -------
        aligned_point : array-like, shape=[..., n_vertices, 3]
            Aligned point.
        """
        discr_geod_path = self.discrete_geodesic_bvp(point, base_point)
        point_ndim_slc = (slice(None),) * self._total_space.point_ndim
        return discr_geod_path[(..., -1) + point_ndim_slc]


class ReparametrizationBundle(FiberBundle):
    """Principal bundle of surfaces module reparametrizations.

    Parameters
    ----------
    total_space : DiscreteSurfaces
        Surfaces equipped with a reparametrization-invariant metric.
    aligner : AlignerAlgorithm
        If None, instantiates default
        RelaxedPathStraightening.
    """

    def __init__(self, total_space, aligner=None):
        if aligner is None:
            aligner = RelaxedPathStraightening(total_space)
        super().__init__(total_space, aligner=aligner)


def _is_iterable(obj):
    """Check if an object is an iterable."""
    return isinstance(obj, (list, tuple))


register_quotient(
    Space=DiscreteSurfaces,
    Metric=ElasticMetric,
    GroupAction="reparametrizations",
    FiberBundle=ReparametrizationBundle,
    QuotientMetric=QuotientMetric,
)



"""Mesh-related data structures."""

import geomstats.backend as gs


class Surface:
    """A surface mesh.

    Mesh info (face centroids, normals and areas) is cached
    for greater performance.

    Parameters
    ----------
    vertices : array-like, shape=[n_vertices, 3]
        Mesh vertices.
    faces : array-like, shape=[n_faces, 3]
        Mesh faces.
    signal : array-like, shape=[n_faces, d]
        A signal of the surface.
    """

    def __init__(self, vertices, faces, signal=None):
        self.vertices = vertices
        self.faces = faces
        self.signal = signal

        (self.face_centroids, self.face_normals, self.face_areas) = (
            self._compute_mesh_info()
        )

    def _compute_mesh_info(self):
        """Compute mesh information.

        Returns
        -------
        face_centroids : array-like, shape=[n_faces, 3]
            Face centroids.
        unit_normals : array-like, shape=[n_faces, 3]
            Face unit normals.
        face_areas : array-like, shape=[n_faces, 1]
            Face areas.
        """
        batch_slc = tuple([slice(None)] * len(self.vertices.shape[:-2]))
        face_coordinates = self.vertices[batch_slc + (self.faces,)]
        vertex_0, vertex_1, vertex_2 = (
            face_coordinates[batch_slc + (slice(None), 0)],
            face_coordinates[batch_slc + (slice(None), 1)],
            face_coordinates[batch_slc + (slice(None), 2)],
        )

        face_centroids = (vertex_0 + vertex_1 + vertex_2) / 3
        normals = 0.5 * gs.cross(vertex_1 - vertex_0, vertex_2 - vertex_0)
        area = gs.linalg.norm(normals, axis=-1)
        unit_normals = gs.einsum("...ij,...i->...ij", normals, 1 / area)

        return face_centroids, unit_normals, gs.expand_dims(area, axis=-1)



from abc import ABC


class PointTransformer(ABC):
    def transform_point(self, point):
        raise NotImplementedError("`transform_point` not implemented")

    def transform_tangent_vec(self, tangent_vec, base_point):
        raise NotImplementedError("`transform_tangent_vec` not implemented")

    def inverse_transform_point(self, other_point):
        raise NotImplementedError("`inverse_transform_point` not implemented")

    def inverse_transform_tangent_vec(self, other_tangent_vec, other_base_point):
        raise NotImplementedError("`inverse_transform_tangent_vec` not implemented")


class IdentityPointTransformer(PointTransformer):
    def transform_point(self, point):
        return point

    def transform_tangent_vec(self, tangent_vec, base_point):
        return tangent_vec

    def inverse_transform_point(self, other_point):
        return other_point

    def inverse_transform_tangent_vec(self, other_tangent_vec, other_base_point):
        return other_tangent_vec


class PointTransformerFromDiffeo(PointTransformer):
    def __init__(self, diffeo):
        self.diffeo = diffeo

    def transform_point(self, point):
        return self.diffeo(point)

    def transform_tangent_vec(self, tangent_vec, base_point):
        return self.diffeo.tangent(tangent_vec, base_point)

    def inverse_transform_point(self, other_point):
        return self.diffeo.inverse(other_point)

    def inverse_transform_tangent_vec(self, other_tangent_vec, other_base_point):
        return self.diffeo.inverse_tangent(other_tangent_vec, other_base_point)



import pytest

import geomstats.backend as gs
from geomstats.geometry.fiber_bundle import FiberBundle
from geomstats.geometry.general_linear import GeneralLinear
from geomstats.geometry.group_action import SpecialOrthogonalComposeAction
from geomstats.geometry.manifold import register_quotient
from geomstats.geometry.matrices import Matrices, MatricesMetric
from geomstats.test.random import RandomDataGenerator
from geomstats.test.test_case import TestCase
from geomstats.test.vectorization import generate_vectorization_data


class FiberBundleTestCase(TestCase):
    tangent_to_multiple = False

    def setup_method(self):
        if not hasattr(self, "data_generator"):
            self.data_generator = RandomDataGenerator(self.total_space)

        if not hasattr(self, "base_data_generator"):
            self.base_data_generator = RandomDataGenerator(self.base)

    def _test_belongs_to_base(self, point, expected, atol):
        res = self.base.belongs(point, atol=atol)
        self.assertAllEqual(res, expected)

    def _test_belongs_to_total_space(self, point, expected, atol):
        res = self.total_space.belongs(point, atol=atol)
        self.assertAllEqual(res, expected)

    def test_riemannian_submersion(self, point, expected, atol):
        res = self.total_space.fiber_bundle.riemannian_submersion(point)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_riemannian_submersion_vec(self, n_reps, atol):
        point = self.data_generator.random_point()
        expected = self.total_space.fiber_bundle.riemannian_submersion(point)

        vec_data = generate_vectorization_data(
            data=[dict(point=point, expected=expected, atol=atol)],
            arg_names=["point"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_riemannian_submersion_belongs_to_base(self, n_points, atol):
        point = self.data_generator.random_point(n_points)

        proj_point = self.total_space.fiber_bundle.riemannian_submersion(point)
        expected = gs.ones(n_points, dtype=bool)

        self._test_belongs_to_base(proj_point, expected, atol)

    def test_lift(self, point, expected, atol):
        res = self.total_space.fiber_bundle.lift(point)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_lift_vec(self, n_reps, atol):
        point = self.base_data_generator.random_point()
        expected = self.total_space.fiber_bundle.lift(point)

        vec_data = generate_vectorization_data(
            data=[dict(point=point, expected=expected, atol=atol)],
            arg_names=["point"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_lift_belongs_to_total_space(self, n_points, atol):
        point = self.base_data_generator.random_point(n_points)
        lifted_point = self.total_space.fiber_bundle.lift(point)

        expected = gs.ones(n_points, dtype=bool)
        self._test_belongs_to_total_space(lifted_point, expected, atol)

    @pytest.mark.random
    def test_riemannian_submersion_after_lift(self, n_points, atol):
        point = self.base_data_generator.random_point(n_points)
        lifted_point = self.total_space.fiber_bundle.lift(point)
        point_ = self.total_space.fiber_bundle.riemannian_submersion(lifted_point)

        self.assertAllClose(point_, point, atol=atol)

    def test_tangent_riemannian_submersion(
        self, tangent_vec, base_point, expected, atol
    ):
        res = self.total_space.fiber_bundle.tangent_riemannian_submersion(
            tangent_vec, base_point
        )
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_tangent_riemannian_submersion_vec(self, n_reps, atol):
        base_point = self.data_generator.random_point()
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        expected = self.total_space.fiber_bundle.tangent_riemannian_submersion(
            tangent_vec, base_point
        )

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    tangent_vec=tangent_vec,
                    base_point=base_point,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["tangent_vec", "base_point"],
            expected_name="expected",
            vectorization_type="sym" if self.tangent_to_multiple else "repeat-0",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_tangent_riemannian_submersion_is_tangent(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        proj_tangent_vector = (
            self.total_space.fiber_bundle.tangent_riemannian_submersion(
                tangent_vec, base_point
            )
        )
        proj_point = self.total_space.fiber_bundle.riemannian_submersion(base_point)

        res = self.base.is_tangent(proj_tangent_vector, proj_point, atol=atol)
        expected = gs.ones(n_points, dtype=bool)
        self.assertAllEqual(res, expected)

    def test_align(self, point, base_point, expected, atol):
        res = self.total_space.fiber_bundle.align(point, base_point)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_align_vec(self, n_reps, atol):
        point = self.data_generator.random_point()
        base_point = self.data_generator.random_point()

        expected = self.total_space.fiber_bundle.align(point, base_point)

        vec_data = generate_vectorization_data(
            data=[
                dict(point=point, base_point=base_point, expected=expected, atol=atol)
            ],
            arg_names=["point", "base_point"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_log_after_align_is_horizontal(self, n_points, atol):
        point = self.data_generator.random_point(n_points)
        base_point = self.data_generator.random_point(n_points)

        aligned_point = self.total_space.fiber_bundle.align(point, base_point)
        log = self.total_space.metric.log(aligned_point, base_point)

        expected = gs.ones(n_points, dtype=bool)
        self.test_is_horizontal(log, base_point, expected, atol)

    def test_horizontal_projection(self, tangent_vec, base_point, expected, atol):
        res = self.total_space.fiber_bundle.horizontal_projection(
            tangent_vec, base_point
        )
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_horizontal_projection_vec(self, n_reps, atol):
        base_point = self.data_generator.random_point()
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        expected = self.total_space.fiber_bundle.horizontal_projection(
            tangent_vec, base_point
        )

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    tangent_vec=tangent_vec,
                    base_point=base_point,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["tangent_vec", "base_point"],
            expected_name="expected",
            vectorization_type="sym" if self.tangent_to_multiple else "repeat-0",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_horizontal_projection_is_horizontal(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        horizontal = self.total_space.fiber_bundle.horizontal_projection(
            tangent_vec, base_point
        )
        expected = gs.ones(n_points, dtype=bool)
        self.test_is_horizontal(horizontal, base_point, expected, atol)

    def test_vertical_projection(self, tangent_vec, base_point, expected, atol):
        res = self.total_space.fiber_bundle.vertical_projection(tangent_vec, base_point)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_vertical_projection_vec(self, n_reps, atol):
        base_point = self.data_generator.random_point()
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        expected = self.total_space.fiber_bundle.vertical_projection(
            tangent_vec, base_point
        )

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    tangent_vec=tangent_vec,
                    base_point=base_point,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["tangent_vec", "base_point"],
            expected_name="expected",
            n_reps=n_reps,
            vectorization_type="sym" if self.tangent_to_multiple else "repeat-0",
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_vertical_projection_is_vertical(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        vertical = self.total_space.fiber_bundle.vertical_projection(
            tangent_vec, base_point
        )
        expected = gs.ones(n_points, dtype=bool)
        self.test_is_vertical(vertical, base_point, expected, atol)

    @pytest.mark.random
    def test_tangent_riemannian_submersion_after_vertical_projection(
        self, n_points, atol
    ):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        vertical = self.total_space.fiber_bundle.vertical_projection(
            tangent_vec, base_point
        )
        res = self.total_space.fiber_bundle.tangent_riemannian_submersion(
            vertical, base_point
        )
        expected = gs.zeros_like(res)

        self.assertAllClose(res, expected, atol=atol)

    def test_is_horizontal(self, tangent_vec, base_point, expected, atol):
        res = self.total_space.fiber_bundle.is_horizontal(
            tangent_vec, base_point, atol=atol
        )
        self.assertAllEqual(res, expected)

    @pytest.mark.vec
    def test_is_horizontal_vec(self, n_reps, atol):
        base_point = self.data_generator.random_point()
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        expected = self.total_space.fiber_bundle.is_horizontal(
            tangent_vec, base_point, atol=atol
        )

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    tangent_vec=tangent_vec,
                    base_point=base_point,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["tangent_vec", "base_point"],
            expected_name="expected",
            vectorization_type="sym" if self.tangent_to_multiple else "repeat-0",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    def test_is_vertical(self, tangent_vec, base_point, expected, atol):
        res = self.total_space.fiber_bundle.is_vertical(
            tangent_vec, base_point, atol=atol
        )
        self.assertAllEqual(res, expected)

    @pytest.mark.vec
    def test_is_vertical_vec(self, n_reps, atol):
        base_point = self.data_generator.random_point()
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        expected = self.total_space.fiber_bundle.is_vertical(
            tangent_vec, base_point, atol=atol
        )

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    tangent_vec=tangent_vec,
                    base_point=base_point,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["tangent_vec", "base_point"],
            expected_name="expected",
            vectorization_type="sym" if self.tangent_to_multiple else "repeat-0",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    def test_horizontal_lift(
        self, tangent_vec, expected, atol, base_point=None, fiber_point=None
    ):
        res = self.total_space.fiber_bundle.horizontal_lift(
            tangent_vec, base_point=base_point, fiber_point=fiber_point
        )
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_horizontal_lift_vec(self, n_reps, atol):
        base_point = self.base_data_generator.random_point()
        tangent_vec = self.base_data_generator.random_tangent_vec(base_point)

        expected = self.total_space.fiber_bundle.horizontal_lift(
            tangent_vec, base_point=base_point
        )

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    tangent_vec=tangent_vec,
                    base_point=base_point,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["tangent_vec", "base_point"],
            expected_name="expected",
            vectorization_type="sym" if self.tangent_to_multiple else "repeat-0",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_horizontal_lift_is_horizontal(self, n_points, atol):
        base_point = self.base_data_generator.random_point(n_points)
        tangent_vec = self.base_data_generator.random_tangent_vec(base_point)

        fiber_point = self.total_space.fiber_bundle.lift(base_point)
        horizontal = self.total_space.fiber_bundle.horizontal_lift(
            tangent_vec, base_point=base_point, fiber_point=fiber_point
        )

        expected = gs.ones(n_points, dtype=bool)
        self.test_is_horizontal(horizontal, fiber_point, expected, atol)

    @pytest.mark.random
    def test_tangent_riemannian_submersion_after_horizontal_lift(self, n_points, atol):
        base_point = self.base_data_generator.random_point(n_points)
        tangent_vec = self.base_data_generator.random_tangent_vec(base_point)
        fiber_point = self.total_space.fiber_bundle.lift(base_point)

        horizontal = self.total_space.fiber_bundle.horizontal_lift(
            tangent_vec, fiber_point=fiber_point
        )
        tangent_vec_ = self.total_space.fiber_bundle.tangent_riemannian_submersion(
            horizontal, fiber_point
        )

        self.assertAllClose(tangent_vec_, tangent_vec, atol=atol)

    def test_integrability_tensor(
        self, tangent_vec_a, tangent_vec_b, base_point, expected, atol
    ):
        res = self.total_space.fiber_bundle.integrability_tensor(
            tangent_vec_a, tangent_vec_b, base_point
        )
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_integrability_tensor_vec(self, n_reps, atol):
        base_point = self.data_generator.random_point()
        tangent_vec_a = self.data_generator.random_tangent_vec(base_point)
        tangent_vec_b = self.data_generator.random_tangent_vec(base_point)

        expected = self.total_space.fiber_bundle.integrability_tensor(
            tangent_vec_a, tangent_vec_b, base_point
        )

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    tangent_vec_a=tangent_vec_a,
                    tangent_vec_b=tangent_vec_b,
                    base_point=base_point,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["tangent_vec_a", "tangent_vec_b", "base_point"],
            expected_name="expected",
            vectorization_type="sym" if self.tangent_to_multiple else "repeat-0-1",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    def test_integrability_tensor_derivative(
        self,
        horizontal_vec_x,
        horizontal_vec_y,
        nabla_x_y,
        tangent_vec_e,
        nabla_x_e,
        base_point,
        expected_nabla_x_a_y_e,
        expected_a_y_e,
        atol,
    ):
        (
            nabla_x_a_y_e,
            a_y_e,
        ) = self.total_space.fiber_bundle.integrability_tensor_derivative(
            horizontal_vec_x,
            horizontal_vec_y,
            nabla_x_y,
            tangent_vec_e,
            nabla_x_e,
            base_point,
        )
        self.assertAllClose(nabla_x_a_y_e, expected_nabla_x_a_y_e, atol=atol)
        self.assertAllClose(a_y_e, expected_a_y_e, atol=atol)

    @pytest.mark.vec
    def test_integrability_tensor_derivative_vec(self, n_reps, atol):
        base_point = self.data_generator.random_point()
        horizontal_vec_x = self.total_space.fiber_bundle.horizontal_lift(
            self.data_generator.random_tangent_vec(base_point),
            fiber_point=base_point,
        )
        horizontal_vec_y = self.total_space.fiber_bundle.horizontal_lift(
            self.data_generator.random_tangent_vec(base_point),
            fiber_point=base_point,
        )
        nabla_x_y = self.data_generator.random_tangent_vec(base_point)
        tangent_vec_e = self.data_generator.random_tangent_vec(base_point)
        nabla_x_e = self.data_generator.random_tangent_vec(base_point)

        (
            nabla_x_a_y_e,
            a_y_e,
        ) = self.total_space.fiber_bundle.integrability_tensor_derivative(
            horizontal_vec_x,
            horizontal_vec_y,
            nabla_x_y,
            tangent_vec_e,
            nabla_x_e,
            base_point,
        )

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    horizontal_vec_x=horizontal_vec_x,
                    horizontal_vec_y=horizontal_vec_y,
                    nabla_x_y=nabla_x_y,
                    tangent_vec_e=tangent_vec_e,
                    nabla_x_e=nabla_x_e,
                    base_point=base_point,
                    expected_nabla_x_a_y_e=nabla_x_a_y_e,
                    expected_a_y_e=a_y_e,
                    atol=atol,
                )
            ],
            arg_names=[
                "horizontal_vec_x",
                "horizontal_vec_y",
                "nabla_x_y",
                "tangent_vec_e",
                "nabla_x_e",
                "base_point",
            ],
            expected_name=["expected_nabla_x_a_y_e", "expected_a_y_e"],
            n_reps=n_reps,
            vectorization_type="basic",
        )
        self._test_vectorization(vec_data)


class GeneralLinearBuresWassersteinBundle(FiberBundle):
    def __init__(self, total_space):
        super().__init__(total_space=total_space, aligner=True)

    @staticmethod
    def riemannian_submersion(point):
        return Matrices.mul(point, Matrices.transpose(point))

    def tangent_riemannian_submersion(self, tangent_vec, base_point):
        product = Matrices.mul(base_point, Matrices.transpose(tangent_vec))
        return 2 * Matrices.to_symmetric(product)

    def horizontal_lift(self, tangent_vec, base_point=None, fiber_point=None):
        if base_point is None and fiber_point is None:
            raise ValueError(
                "Either a point (of the total space) or a "
                "base point (of the base manifold) must be "
                "given."
            )

        if base_point is None:
            base_point = self.riemannian_submersion(fiber_point)

        if fiber_point is None:
            fiber_point = self.lift(base_point)

        sylvester = gs.linalg.solve_sylvester(base_point, base_point, tangent_vec)
        return Matrices.mul(sylvester, fiber_point)

    @staticmethod
    def lift(point):
        return gs.linalg.cholesky(point)


register_quotient(
    Space=GeneralLinear,
    Metric=MatricesMetric,
    GroupAction=SpecialOrthogonalComposeAction,
    FiberBundle=GeneralLinearBuresWassersteinBundle,
)



"""Statistical Manifold of geometric distributions with the Fisher metric.

Lead author: Tra My Nguyen.
"""

from scipy.stats import geom

import geomstats.backend as gs
from geomstats.geometry.base import VectorSpaceOpenSet
from geomstats.geometry.euclidean import Euclidean
from geomstats.geometry.riemannian_metric import RiemannianMetric
from geomstats.information_geometry.base import (
    InformationManifoldMixin,
    ScipyUnivariateRandomVariable,
)


class GeometricDistributions(InformationManifoldMixin, VectorSpaceOpenSet):
    """Class for the manifold of geometric distributions.

    This is the parameter space of geometric distributions
    i.e. [0,1] segment.
    """

    def __init__(self, equip=True):
        super().__init__(
            dim=1,
            embedding_space=Euclidean(dim=1),
            support_shape=(),
            equip=equip,
        )
        self._scp_rv = GeometricDistributionsRandomVariable(self)

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return GeometricMetric

    def belongs(self, point, atol=gs.atol):
        """Evaluate if a point belongs to the manifold of geometric distributions.

        Parameters
        ----------
        point : array-like, shape=[..., 1]
            Point to be checked.
        atol : float
            Tolerance to evaluate positivity.
            Optional, default: gs.atol

        Returns
        -------
        belongs : array-like, shape=[...,]
            Boolean indicating whether point represents an geometric
            distribution.
        """
        belongs_shape = self.shape == point.shape[-self.point_ndim :]
        if not belongs_shape:
            shape = point.shape[: -self.point_ndim]
            return gs.zeros(shape, dtype=bool)
        return gs.squeeze(gs.logical_and(-atol <= point, point <= 1 + atol))

    def random_point(self, n_samples=1):
        """Sample parameters of geometric distributions.

        The uniform distribution on (0, 1) is used.

        Parameters
        ----------
        n_samples : int
            Number of samples.
            Optional, default: 1.
        bound : float
            Right-end ot the segment where geometric parameters are sampled.
            Optional, default: 1.

        Returns
        -------
        samples : array-like, shape=[n_samples,]
            Sample of points representing geometric distributions.
        """
        size = (n_samples, self.dim) if n_samples != 1 else (self.dim,)
        return gs.random.rand(*size)

    def projection(self, point, atol=gs.atol):
        """Project a point in ambient space to the open set.

        Return a new point belonging to [0,1] segment within the given tolerance.

        Parameters
        ----------
        point : array-like, shape=[..., 1]
            Point in ambient space.
        atol : float
            Tolerance to evaluate the position with respect to [0,1] interval.

        Returns
        -------
        projected : array-like, shape=[..., 1]
            Projected point.
        """
        return gs.where(
            gs.logical_or(point < atol, point > 1 - atol),
            (1 - atol) * gs.cast(point > 1 - atol, point.dtype)
            + atol * gs.cast(point < atol, point.dtype),
            point,
        )

    def sample(self, point, n_samples=1):
        """Sample from the geometric distribution.

        Sample from the geometric distribution with parameter provided
        by point.

        Parameters
        ----------
        point : array-like, shape=[..., 1]
            Point representing an geometric distribution.
        n_samples : int
            Number of points to sample with each parameter in point.
            Optional, default: 1.

        Returns
        -------
        samples : array-like, shape=[..., n_samples]
            Sample from geometric distributions.
        """
        return self._scp_rv.rvs(point, n_samples)

    def point_to_pdf(self, point):
        """Compute pmf associated to point.

        Compute the probability mass function of the geometric
        distribution with parameters provided by point.

        Parameters
        ----------
        point : array-like, shape=[..., 1]
            Point representing an geometric distribution (scale).

        Returns
        -------
        pdf : function
            Probability mass function of the geometric distribution with
            scale parameter provided by point.
        """

        def pmf(k):
            """Generate parameterized function for geometric pmf.

            Compute the probability mass function of the geometric
            distribution with parameters provided by point.

            Parameters
            ----------
            k : array-like, shape=[n_samples,]
                Point in the sample space, i.e. an integer.

            Returns
            -------
            pmf_at_k : array-like, shape=[..., n_samples]
                Probability mass function of the geometric distribution with
                parameters provided by point.
            """
            k = gs.reshape(gs.array(k), (-1,))
            return (1 - point) ** (k - 1) * point

        return pmf


class GeometricMetric(RiemannianMetric):
    """Class for the Fisher information metric on geometric distributions.

    References
    ----------
    .. [AM1981] Atkinson, C., & Mitchell, A. F. (1981). Rao's distance measure.
        Sankhyā: The Indian Journal of Statistics, Series A, 345-365.
    """

    def squared_dist(self, point_a, point_b):
        """Compute squared distance associated with the geometric metric.

        Parameters
        ----------
        point_a : array-like, shape=[..., 1]
            Point representing an geometric distribution.
        point_b : array-like, shape=[..., 1]
            Point representing a geometric distribution.

        Returns
        -------
        squared_dist : array-like, shape=[...,]
            Squared distance between points point_a and point_b.
        """
        return gs.squeeze(
            4
            * (gs.arctanh(gs.sqrt(1 - point_a)) - gs.arctanh(gs.sqrt(1 - point_b))) ** 2
        )

    def metric_matrix(self, base_point):
        """Compute the metric matrix at base_point.

        Parameters
        ----------
        base_point : array-like, shape=[..., 1]
            Point representing a geometric distribution.

        Returns
        -------
        mat : array-like, shape=[..., 1, 1]
            Metric matrix.
        """
        return gs.expand_dims(1 / (base_point**2 * (1 - base_point)), axis=-1)

    @staticmethod
    def _geodesic_path(t, frequency, initial_phase):
        """Generate parameterized function for geodesic curve.

        Parameters
        ----------
        t : array-like, shape=[n_times,]
            Times at which to compute points of the geodesics.

        Returns
        -------
        geodesic : array-like, shape=[..., n_times, 1]
            Values of the geodesic at times t.
        """
        return gs.expand_dims(1 - gs.tanh(frequency * t + initial_phase) ** 2, axis=-1)

    def _geodesic_ivp(self, initial_point, initial_tangent_vec):
        """Solve geodesic initial value problem.

        Compute the parameterized function for the geodesic starting at
        initial_point with initial velocity given by initial_tangent_vec.

        Parameters
        ----------
        initial_point : array-like, shape=[..., 1]
            Initial point.

        initial_tangent_vec : array-like, shape=[..., 1]
            Tangent vector at initial point.

        Returns
        -------
        path : function
            Parameterized function for the geodesic curve starting at
            initial_point with velocity initial_tangent_vec.
        """
        initial_phase = gs.arctanh(gs.sqrt(1 - initial_point))
        frequency = -initial_tangent_vec / (
            2 * initial_point * gs.sqrt(1 - initial_point)
        )
        return lambda t: self._geodesic_path(t, frequency, initial_phase)

    def _geodesic_bvp(self, initial_point, end_point):
        """Solve geodesic boundary problem.

        Compute the parameterized function for the geodesic starting at
        initial_point and ending at end_point.

        Parameters
        ----------
        initial_point : array-like, shape=[..., 1]
            Initial point.
        end_point : array-like, shape=[..., 1]
            End point.

        Returns
        -------
        path : function
            Parameterized function for the geodesic curve starting at
            initial_point and ending at end_point.
        """
        initial_phase = gs.arctanh(gs.sqrt(1 - initial_point))
        frequency = gs.arctanh(gs.sqrt(1 - end_point)) - initial_phase
        return lambda t: self._geodesic_path(t, frequency, initial_phase)

    def exp(self, tangent_vec, base_point):
        """Compute exp map of a base point in tangent vector direction.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., 1]
            Tangent vector at base point.
        base_point : array-like, shape=[..., 1]
            Base point.

        Returns
        -------
        exp : array-like, shape=[..., 1]
            End point of the geodesic starting at base_point with
            initial velocity tangent_vec.
        """
        return (
            1
            - gs.tanh(
                -tangent_vec / (2 * base_point * gs.sqrt(1 - base_point))
                + gs.arctanh(gs.sqrt(1 - base_point))
            )
            ** 2
        )

    def log(self, point, base_point):
        """Compute log map using a base point and an end point.

        Parameters
        ----------
        point : array-like, shape=[..., 1]
            End point.
        base_point : array-like, shape=[..., 1]
            Base point.

        Returns
        -------
        tangent_vec : array-like, shape=[..., 1]
            Initial velocity of the geodesic starting at base_point and
            reaching point at time 1.
        """
        return (
            -2
            * base_point
            * gs.sqrt(1 - base_point)
            * (gs.arctanh(gs.sqrt(1 - point)) - gs.arctanh(gs.sqrt(1 - base_point)))
        )


class GeometricDistributionsRandomVariable(ScipyUnivariateRandomVariable):
    """A geometric random variable."""

    def __init__(self, space):
        super().__init__(space, geom.rvs)

    @staticmethod
    def _flatten_params(point, pre_flat_shape):
        flat_point = gs.reshape(gs.broadcast_to(point, pre_flat_shape), (-1,))
        return {"p": flat_point}

    def pdf(self, x, point):
        """Evaluate the probability density function at x.

        Parameters
        ----------
        x : array-like, shape=[n_samples, *support_shape]
            Sample points at which to compute the probability density function.
        point : array-like, shape=[..., *space.shape]
            Point representing a distribution.

        Returns
        -------
        pdf_at_x : array-like, shape=[..., n_samples, *support_shape]
            Values of pdf at x for each value of the parameters provided
            by point.
        """
        return gs.from_numpy(
            geom.pmf(x, point),
        )



"""Decorator to handle vectorization.

This abstracts the backend type.
"""

import math

import geomstats.backend as gs


def _get_max_ndim_point(*point):
    """Identify point with higher dimension.

    Parameters
    ----------
    point : array-like

    Returns
    -------
    max_ndim_point : array-like
        Point with higher dimension.
    """
    max_ndim_point = point[0]
    for point_ in point[1:]:
        if point_.ndim > max_ndim_point.ndim:
            max_ndim_point = point_

    return max_ndim_point


def get_n_points(point_ndim, *point):
    """Compute the number of points.

    Parameters
    ----------
    point_ndim : int
        Point number of array dimensions.
    point : array-like
        Point belonging to the space.

    Returns
    -------
    n_points : int
        Number of points.
    """
    point_max_ndim = _get_max_ndim_point(*point)
    return math.prod(point_max_ndim.shape[:-point_ndim])


def check_is_batch(point_ndim, *point):
    """Check if inputs are batch.

    Parameters
    ----------
    point_ndim : int
        Point number of array dimensions.
    point : array-like
        Point belonging to the space.

    Returns
    -------
    is_batch : bool
        Returns True if point contains several points.
    """
    return any(point_.ndim > point_ndim for point_ in point)


def get_batch_shape(point_ndim, *point):
    """Get batch shape.

    Parameters
    ----------
    point_ndim : int
        Point number of array dimensions.
    point : array-like or None
        Point belonging to the space.

    Returns
    -------
    batch_shape : tuple
        Returns the shape related with batch. () if only one point.
    """
    point = list(filter(_is_not_none, point))
    if len(point) == 0:
        return ()
    point_max_ndim = _get_max_ndim_point(*point)
    return point_max_ndim.shape[:-point_ndim]


def repeat_point(point, n_reps=2, expand=False):
    """Repeat point.

    Parameters
    ----------
    point : array-like
        Point of a space.
    n_reps : int
        Number of times the point should be repeated.
    expand : bool
        Repeat even if n_reps == 1.

    Returns
    -------
    rep_point : array-like
        point repeated n_reps times.
    """
    if not expand and n_reps == 1:
        return gs.copy(point)

    return gs.repeat(gs.expand_dims(point, 0), n_reps, axis=0)


def _is_not_none(value):
    """Check if a value is None."""
    return value is not None


def repeat_out(point_ndim, out, *point, out_shape=()):
    """Repeat out shape after finding batch shape.

    Parameters
    ----------
    point_ndim : int
        Point number of array dimensions.
    out : array-like
        Output to be repeated
    point : array-like or None
        Point belonging to the space.
    out_shape : tuple
        Indicates out shape for no batch computations.

    Returns
    -------
    out : array-like
        If no batch, then input is returned. Otherwise it is broadcasted.
    """
    point = filter(_is_not_none, point)
    batch_shape = get_batch_shape(point_ndim, *point)
    if out.shape[: -len(out_shape)] != batch_shape:
        return gs.broadcast_to(out, batch_shape + out_shape)
    return out


def repeat_out_multiple_ndim(
    out, point_ndim_1, points_1, point_ndim_2, points_2, out_ndim=0
):
    """Repeat out after finding batch shape.

    Differs from `repeat_out` by accepting two sets of point_ndim arrays.

    Parameters
    ----------
    out : array-like
        Output to be repeated
    point_ndim_1 : int
        Point number of array dimensions.
    points_1 : tuple[array-like or None]
        Arrays of dimension point_ndim_1 or higher.
    point_ndim_2 : int
        Point number of array dimensions.
    points_2 : tuple[array-like or None]
        Arrays of dimension point_ndim_2 or higher.
    out_ndim : int
        Out number of array dimensions.

    Returns
    -------
    out : array-like
        If no batch, then input is returned. Otherwise it is broadcasted.
    """
    batch_shape = get_batch_shape(point_ndim_1, *points_1)
    if not batch_shape:
        batch_shape = get_batch_shape(point_ndim_2, *points_2)

    out_shape = out.shape[-out_ndim:]
    if out.shape[:-out_ndim] != batch_shape:
        return gs.broadcast_to(out, batch_shape + out_shape)
    return out


def broadcast_to_multibatch(batch_shape_a, batch_shape_b, array_a, *array_b):
    """Broadcast to multibatch.

    Gives to both arrays batch shape `batch_shape_b + batch_shape_a`.

    Does nothing if one of the batch shapes is empty.

    Parameters
    ----------
    batch_shape_a : tuple
        Batch shape of array_a.
    batch_shape_b : tuple
        Batch shape of array_b.
    array_a : array
    array_b : array
    """
    multi_b = len(array_b) > 1

    if not batch_shape_a or not batch_shape_b:
        return (array_a, array_b) if multi_b else (array_a, array_b[0])

    array_a_ = gs.broadcast_to(array_a, batch_shape_b + array_a.shape)

    n_batch_b = len(batch_shape_b)
    indices_in = list(range(len(batch_shape_a)))
    indices_out = [index + n_batch_b for index in indices_in]

    array_b_ = []
    for array in array_b:
        array_b_aux = gs.broadcast_to(array, batch_shape_a + array.shape)
        array_b_.append(gs.moveaxis(array_b_aux, indices_in, indices_out))

    return (array_a_, array_b_) if multi_b else (array_a_, array_b_[0])



import pytest

import geomstats.backend as gs
from geomstats.test.random import RandomDataGenerator
from geomstats.test.test_case import TestCase
from geomstats.vectorization import repeat_point


class BaseEstimatorTestCase(TestCase):
    def setup_method(self):
        if not hasattr(self, "data_generator"):
            self.data_generator = RandomDataGenerator(self.estimator.space)


class MeanEstimatorMixinsTestCase:
    def _self_assert_same_point(self, point, point_, atol):
        self.assertAllClose(point, point_, atol)

    def test_fit(self, X, expected, atol, weights=None):
        kwargs = {} if weights is None else {"weights": weights}
        res = self.estimator.fit(X, **kwargs).estimate_
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.random
    def test_one_point(self, atol):
        X = gs.expand_dims(self.data_generator.random_point(n_points=1), axis=0)

        mean = self.estimator.fit(X).estimate_
        self._self_assert_same_point(mean, X[0], atol)

    @pytest.mark.random
    def test_n_times_same_point(self, n_samples, atol):
        X = repeat_point(self.data_generator.random_point(n_points=1), n_samples)

        mean = self.estimator.fit(X).estimate_
        self._self_assert_same_point(mean, X[0], atol)

    @pytest.mark.random
    def test_estimate_belongs(self, n_samples, atol):
        X = self.data_generator.random_point(n_points=n_samples)
        mean = self.estimator.fit(X).estimate_
        belongs = self.estimator.space.belongs(mean, atol=atol)
        self.assertTrue(belongs)


class ClusterMixinsTestCase:
    @pytest.mark.random
    def test_n_repeated_clusters(self, n_samples, atol):
        n_clusters = self.estimator.n_clusters
        X = []
        for _ in range(n_clusters):
            X.extend(
                repeat_point(self.data_generator.random_point(n_points=1), n_samples)
            )
        X = gs.array(X)

        cluster_centers = self.estimator.fit(X).cluster_centers_
        labels = self.estimator.predict(X)

        for k in range(n_clusters):
            i = n_samples * k
            j = i + n_samples
            X_ = X[i:j]
            labels_ = labels[i:j]

            self.assertTrue(gs.unique(labels_).shape[0], 1)
            cluster_center = cluster_centers[labels_[0]]

            dist = self.estimator.space.metric.dist(X_, cluster_center)
            self.assertAllClose(dist, gs.zeros(n_samples), atol=atol)

    @pytest.mark.random
    def test_cluster_assignment(self, n_samples):
        X = self.data_generator.random_point(n_samples)
        space = self.estimator.space

        cluster_centers = self.estimator.fit(X).cluster_centers_
        result = self.estimator.predict(X)

        expected = gs.array(
            [
                int(space.metric.closest_neighbor_index(x_i, cluster_centers))
                for x_i in X
            ]
        )
        self.assertAllEqual(result, expected)

    @pytest.mark.random
    def test_cluster_centers_belong(self, n_samples):
        X = self.data_generator.random_point(n_samples)
        space = self.estimator.space

        cluster_centers = self.estimator.fit(X).cluster_centers_

        result = space.belongs(cluster_centers)
        expected = gs.ones((self.estimator.n_clusters,), dtype=bool)
        self.assertAllEqual(result, expected)

    @pytest.mark.shape
    def test_cluster_centers_shape(self, n_samples):
        X = self.data_generator.random_point(n_samples)
        space = self.estimator.space

        cluster_centers = self.estimator.fit(X).cluster_centers_

        n_clusters = self.estimator.n_clusters

        self.assertAllEqual(gs.shape(cluster_centers), (n_clusters, *space.shape))



"""Expectation maximization algorithm.

Lead authors: Thomas Gerald and Hadi Zaatiti.
"""

import logging

from sklearn.base import BaseEstimator, ClusterMixin

import geomstats.backend as gs
from geomstats.learning._template import TransformerMixin
from geomstats.learning.frechet_mean import FrechetMean, variance
from geomstats.learning.kmeans import RiemannianKMeans

PDF_TOL = 1e-6
SUM_CHECK_PDF = 1e-4
MIN_VAR_INIT = 1e-3


class GaussianMixtureModel:
    r"""Gaussian mixture model (GMM).

    Parameters
    ----------
    space : Manifold
        Equipped manifold.
    means : array-like, shape=[n_gaussians, dim]
        Means of each component of the GMM.
    variances : array-like, shape=[n_gaussians,]
        Variances of each component of the GMM.

    Attributes
    ----------
    normalization_factor_var : array-like, shape=[n_variances,]
        Array of computed normalization factor.
    variances_range : array-like, shape=[n_variances,]
        Array of standard deviations.
    phi_inv_var : array-like, shape=[n_variances,]
        Array of the computed inverse of a function phi
        whose expression is closed-form
        :math:`\sigma\mapsto \sigma^3 \times \frac{d}
        {\mathstrut d\sigma}\log \zeta_m(\sigma)`
        where :math:`\sigma` denotes the variance
        and :math:`\zeta` the normalization coefficient
        and :math:`m` the dimension.
    """

    def __init__(
        self,
        space,
        means=None,
        variances=None,
        zeta_lower_bound=5e-1,
        zeta_upper_bound=2.0,
        zeta_step=0.01,
    ):
        self.space = space
        self.means = means
        self.variances = variances

        self.zeta_lower_bound = zeta_lower_bound
        self.zeta_upper_bound = zeta_upper_bound
        self.zeta_step = zeta_step

        (
            self.variances_range,
            self.normalization_factor_var,
            self.phi_inv_var,
        ) = self._normalization_factor_init()

    def _normalization_factor_init(self):
        r"""Set up function for the normalization factor.

        The normalization factor is used to define Gaussian distributions
        at initialization.
        """
        variances = gs.arange(
            self.zeta_lower_bound, self.zeta_upper_bound, self.zeta_step
        )
        normalization_factor_var = self.space.metric.normalization_factor(variances)

        cond_1 = normalization_factor_var.sum() != normalization_factor_var.sum()
        cond_2 = normalization_factor_var.sum() == float("+inf")
        cond_3 = normalization_factor_var.sum() == float("-inf")

        if cond_1 or cond_2 or cond_3:
            logging.warning("Untractable normalization factor :")

            limit_nf = (
                ((normalization_factor_var / normalization_factor_var) * 0)
                .nonzero()[0]
                .item()
            )
            max_nf = len(variances)
            variances = variances[0:limit_nf]
            normalization_factor_var = normalization_factor_var[0:limit_nf]
            if cond_1:
                logging.warning("\t Nan value " "in processing normalization factor")
            if cond_2 or cond_3:
                raise ValueError("\t +-inf value in " "processing normalization factor")

            logging.warning("\t Max variance is now : %s", str(variances[-1]))
            logging.warning(
                "\t Number of possible variance is now: %s / %s ",
                str(len(variances)),
                str(max_nf),
            )

        _, log_grad_zeta = self.space.metric.norm_factor_gradient(variances)

        phi_inv_var = variances**3 * log_grad_zeta

        return variances, normalization_factor_var, phi_inv_var

    def pdf(self, data):
        """Return the separate probability density function of GMM.

        The probability density function is computed for
        each component of the GMM separately (i.e., mixture coefficients
        are not taken into account).

        Parameters
        ----------
        data : array-like, shape=[n_samples, dim]
            Points at which the GMM probability density is computed.

        Returns
        -------
        pdf : array-like, shape=[n_samples, n_gaussians,]
            Probability density function computed at each data
            sample and for each component of the GMM.
        """
        data_length, _, _ = data.shape + (self.means.shape[0],)

        variances_expanded = gs.expand_dims(self.variances, 0)
        variances_expanded = gs.repeat(variances_expanded, data_length, 0)

        variances_flatten = variances_expanded.flatten()

        distances = -(self.space.metric.dist_broadcast(data, self.means) ** 2)
        distances = gs.reshape(distances, (data.shape[0] * self.variances.shape[0],))

        num = gs.exp(distances / (2 * variances_flatten**2))

        den = self._compute_normalization_factor()

        den = gs.expand_dims(den, 0)
        den = gs.repeat(den, data_length, axis=0).flatten()

        pdf = num / den
        pdf = gs.reshape(pdf, (data.shape[0], self.means.shape[0]))

        return pdf

    def _compute_normalization_factor(self):
        """Find the normalization factor given some variances.

        Returns
        -------
        norm_factor : array-like, shape=[n_gaussians,]
            Array of normalization factors for the given
            variances.
        """
        n_gaussians, precision = self.variances.shape[0], self.variances_range.shape[0]

        ref = gs.expand_dims(self.variances_range, 0)
        ref = gs.repeat(ref, n_gaussians, axis=0)
        val = gs.expand_dims(self.variances, 1)
        val = gs.repeat(val, precision, axis=1)

        difference = gs.abs(ref - val)

        index = gs.argmin(difference, axis=-1)
        norm_factor = self.normalization_factor_var[index]

        return norm_factor

    def compute_variance_from_index(self, weighted_distances):
        r"""Return the variance given weighted distances.

        Parameters
        ----------
        weighted_distances : array-like, shape=[n_gaussians,]
            Mean of the weighted distances between training data
            and current barycentres. The weights of each data sample
            corresponds to the probability of belonging to a component
            of the Gaussian mixture model.

        Returns
        -------
        var : array-like, shape=[n_gaussians,]
            Estimated variances for each component of the GMM.
        """
        n_gaussians, precision = (
            weighted_distances.shape[0],
            self.variances_range.shape[0],
        )

        ref = gs.expand_dims(self.phi_inv_var, 0)
        ref = gs.repeat(ref, n_gaussians, axis=0)

        val = gs.expand_dims(weighted_distances, 1)
        val = gs.repeat(val, precision, axis=1)

        abs_difference = gs.abs(ref - val)

        index = gs.argmin(abs_difference, -1)

        var = self.variances_range[index]

        return var

    def weighted_pdf(self, mixture_coefficients, mesh_data):
        """Return the probability density function of a GMM.

        Parameters
        ----------
        mixture_coefficients : array-like, shape=[n_gaussians,]
            Coefficients of the Gaussian mixture model.
        mesh_data : array-like, shape=[n_precision, dim]
            Points at which the GMM probability density is computed.

        Returns
        -------
        weighted_pdf : array-like, shape=[n_precision, n_gaussians,]
            Probability density function computed for each point of
            the mesh data, for each component of the GMM.
        """
        distance_to_mean = self.space.metric.dist_broadcast(mesh_data, self.means)

        variances_units = gs.expand_dims(self.variances, 0)
        variances_units = gs.repeat(variances_units, distance_to_mean.shape[0], axis=0)

        distribution_normal = gs.exp(-(distance_to_mean**2) / (2 * variances_units**2))

        zeta_sigma = (2 * gs.pi) ** (2 / 3) * self.variances
        zeta_sigma = zeta_sigma * gs.exp(
            (self.variances**2 / 2) * gs.erf(self.variances / gs.sqrt(2))
        )

        result_num = gs.expand_dims(mixture_coefficients, 0)
        result_num = gs.repeat(result_num, len(distribution_normal), axis=0)
        result_num = result_num * distribution_normal

        result_denum = gs.expand_dims(zeta_sigma, 0)
        result_denum = gs.repeat(result_denum, len(distribution_normal), axis=0)

        weighted_pdf = result_num / result_denum

        return weighted_pdf


class RiemannianEM(TransformerMixin, ClusterMixin, BaseEstimator):
    r"""Expectation-maximization algorithm.

    A class for performing Expectation-Maximization to fit a Gaussian Mixture
    Model (GMM) to data on a manifold. This method is only implemented for
    the hypersphere and the Poincare ball.

    Parameters
    ----------
    space : Manifold
        Equipped manifold.
    n_gaussians : int
        Number of Gaussian components in the mix.
    initialisation_method : basestring
        Optional, default: 'random'.
        Choice between initialization method for variances, means and weights.

        - 'random' : will select random uniformly train points as
          initial cluster centers.
        - 'kmeans' : will apply Riemannian kmeans to deduce
          variances and means that the EM will use initially.
    tol : float
        Optional, default: 1e-2.
        Convergence tolerance. If the difference of mean distance
        between two steps is lower than tol.
    max_iter : int
        Maximum number of iterations for the gradient descent.
        Optional, default: 100.

    Attributes
    ----------
    mixture_coefficients_ : array-like, shape=[n_gaussians,]
        Weights for each GMM component.
    variances_ : array-like, shape=[n_gaussians,]
        Variances for each GMM component.
    means_ : array-like, shape=[n_gaussian, _dimension]
        Barycentre of each component of the GMM.

    Example
    -------
    Available example on the Poincaré Ball manifold
    :mod:`examples.plot_expectation_maximization_ball`
    """

    def __init__(
        self,
        space,
        n_gaussians=8,
        initialisation_method="random",
        tol=1e-2,
        max_iter=100,
        conv_rate=1e-4,
        minimum_epochs=10,
    ):
        self.space = space
        self.n_gaussians = n_gaussians
        self.initialisation_method = initialisation_method
        self.tol = tol
        self.max_iter = max_iter
        self.conv_rate = conv_rate
        self.minimum_epochs = minimum_epochs

        self.mean_estimator = FrechetMean(space)
        if isinstance(self.mean_estimator, FrechetMean):
            self.mean_estimator.method = "batch"
            self.mean_estimator.set(
                max_iter=100,
                epsilon=1e-4,
                init_step_size=1.0,
            )

        self._model = GaussianMixtureModel(self.space)

        self.mixture_coefficients_ = None

    @property
    def means_(self):
        """Means of each component of the GMM."""
        return self._model.means

    @property
    def variances_(self):
        """Array of standard deviations."""
        return self._model.variances

    def _update_posterior_probabilities(self, posterior_probabilities):
        """Posterior probabilities update function.

        Parameters
        ----------
        posterior_probabilities : array-like, shape=[n_samples, n_gaussians,]
            Probability of a given sample to belong to a component
            of the GMM, computed for all components.
        """
        self.mixture_coefficients_ = gs.mean(posterior_probabilities, 0)

        if gs.any(gs.isnan(self.mixture_coefficients_)):
            logging.warning(
                "UPDATE : mixture coefficients contain elements that are not numbers"
            )

    def _update_means(self, data, posterior_probabilities):
        """Update means."""
        n_gaussians = posterior_probabilities.shape[-1]

        data_expand = gs.expand_dims(data, 1)
        data_expand = gs.repeat(data_expand, n_gaussians, axis=1)

        self.mean_estimator.fit(data_expand, weights=posterior_probabilities)
        self._model.means = gs.squeeze(self.mean_estimator.estimate_)

        if gs.any(gs.isnan(self._model.means)):
            logging.warning("UPDATE : means contain not a number elements")

    def _update_variances(self, data, posterior_probabilities):
        """Update variances function.

        Parameters
        ----------
        data : array-like, shape=[n_samples, n_features,]
            Training data, where n_samples is the number of samples and
            n_features is the number of features.
        posterior_probabilities : array-like, shape=[n_samples, n_gaussians,]
            Probability of a given sample to belong to a component
            of the GMM, computed for all components.
        """
        dist_means_data = self.space.metric.dist_broadcast(data, self._model.means) ** 2

        weighted_dist_means_data = (dist_means_data * posterior_probabilities).sum(
            0
        ) / posterior_probabilities.sum(0)

        self._model.variances = self._model.compute_variance_from_index(
            weighted_dist_means_data
        )

        if gs.any(gs.isnan(self._model.variances)):
            logging.warning("UPDATE : variances contain not a number elements")

    def _expectation(self, data):
        """Update the posterior probabilities.

        Parameters
        ----------
        data : array-like, shape=[n_samples, n_features]
            Training data, where n_samples is the number of samples and
            n_features is the number of features.
        """
        pdf = self._model.pdf(data)

        if gs.any(gs.isnan(pdf)):
            logging.warning(
                "EXPECTATION : Probability distribution function"
                "contain elements that are not numbers"
            )

        num_normalized_pdf = gs.einsum("j,...j->...j", self.mixture_coefficients_, pdf)
        valid_pdf_condition = gs.amin(gs.sum(num_normalized_pdf, -1))

        if valid_pdf_condition <= PDF_TOL:
            num_normalized_pdf[gs.sum(num_normalized_pdf, -1) <= PDF_TOL] = 1

        sum_pdf = gs.sum(num_normalized_pdf, -1)
        posterior_probabilities = gs.einsum(
            "...i,...->...i", num_normalized_pdf, 1 / sum_pdf
        )

        if gs.any(gs.mean(posterior_probabilities)) is None:
            logging.warning(
                "EXPECTATION : posterior probabilities "
                "contain elements that are not numbers."
            )

        if (
            1 - SUM_CHECK_PDF
            >= gs.mean(gs.sum(posterior_probabilities, 1))
            >= 1 + SUM_CHECK_PDF
        ):
            logging.warning("EXPECTATION : posterior probabilities " "do not sum to 1.")

        if gs.any(gs.sum(posterior_probabilities, 0) < PDF_TOL):
            logging.warning(
                "EXPECTATION : Gaussian got no elements "
                "(precision error) reinitialize"
            )
            posterior_probabilities[posterior_probabilities == 0] = PDF_TOL

        return posterior_probabilities

    def _maximization(self, data, posterior_probabilities):
        """Update function for the means and variances.

        Parameters
        ----------
        data : array-like, shape=[n_samples, n_features,]
            Training data, where n_samples is the number of samples and
            n_features is the number of features.
        posterior_probabilities : array-like, shape=[n_samples, n_gaussians,]
            Probability of a given sample to belong to a component
            of the GMM, computed for all components.
        """
        self._update_posterior_probabilities(posterior_probabilities)
        self._update_means(data, posterior_probabilities)
        self._update_variances(data, posterior_probabilities)

    def _initialization(self, X):
        if self.initialisation_method == "kmeans":
            kmeans_estimator = RiemannianKMeans(
                space=self.space,
                n_clusters=self.n_gaussians,
                init="random",
            )

            kmeans_estimator.fit(X=X)
            cluster_centers = kmeans_estimator.cluster_centers_
            labels = kmeans_estimator.labels_

            means = cluster_centers
            variances = gs.zeros(self.n_gaussians)

            labeled_data = gs.vstack([labels, gs.transpose(X)])
            labeled_data = gs.transpose(labeled_data)
            for label, cluster_center in enumerate(cluster_centers):
                label_mask = gs.where(labeled_data[:, 0] == label)
                grouped_by_label = labeled_data[label_mask][:, 1:]
                v = variance(self.space, grouped_by_label, cluster_center)
                if grouped_by_label.shape[0] == 1:
                    v += MIN_VAR_INIT
                variances[label] = v
        else:
            dim = self.space.shape[-1]
            means = (gs.random.rand(self.n_gaussians, dim) - 0.5) / dim
            variances = gs.random.rand(self.n_gaussians) / 10 + 0.8

        return means, variances

    def fit(self, X, y=None):
        """Fit a Gaussian mixture model (GMM) given the data.

        Alternates between Expectation and Maximization steps
        for some number of iterations.

        Parameters
        ----------
        X : array-like, shape=[n_samples, n_features]
            Training data, where n_samples is the number of samples
            and n_features is the number of features.
        y : None
            Target values. Ignored.

        Returns
        -------
        self : object
            Returns self.
        """
        self._model.means, self._model.variances = self._initialization(X)

        self.mixture_coefficients_ = gs.ones(self.n_gaussians) / self.n_gaussians
        posterior_probabilities = gs.ones((X.shape[0], self.n_gaussians))

        for epoch in range(self.max_iter):
            old_posterior_probabilities = posterior_probabilities

            posterior_probabilities = self._expectation(X)

            condition = gs.mean(
                gs.abs(old_posterior_probabilities - posterior_probabilities)
            )

            if condition < self.conv_rate and epoch > self.minimum_epochs:
                logging.info("EM converged in %s iterations", epoch)
                break

            self._maximization(X, posterior_probabilities)
        else:
            logging.info(
                "WARNING: EM did not converge \nPlease increase MINIMUM_EPOCHS."
            )

        return self



"""Classes for the pullback metric.

Lead author: Nina Miolane.
"""

import geomstats.backend as gs
from geomstats.geometry.riemannian_metric import RiemannianMetric
from geomstats.numerics.geodesic import ExpODESolver, LogShootingSolver
from geomstats.numerics.ivp import GSIVPIntegrator
from geomstats.vectorization import check_is_batch


class PullbackMetric(RiemannianMetric):
    r"""Pullback metric.

    Let :math:`f` be an immersion :math:`f: M \rightarrow N`
    of one manifold :math:`M` into the Riemannian manifold :math:`N`
    with metric :math:`g`.
    The pull-back metric :math:`f^*g` is defined on :math:`M` for a
    base point :math:`p` as:
    :math:`(f^*g)_p(u, v) = g_{f(p)}(df_p u , df_p v)
    \quad \forall u, v \in T_pM`

    Note
    ----
    The pull-back metric is currently only implemented for an
    immersion into the Euclidean space, i.e. for

    :math:`N=\mathbb{R}^n`.
    """

    def __init__(self, space, signature=None):
        super().__init__(space, signature)
        self._instantiate_solvers()

    def _instantiate_solvers(self):
        if gs.has_autodiff():
            self.log_solver = LogShootingSolver(self._space)

        self.exp_solver = ExpODESolver(
            self._space, integrator=GSIVPIntegrator(n_steps=100, step_type="euler")
        )

    def metric_matrix(self, base_point):
        r"""Metric matrix at the tangent space at a base point.

        Let :math:`f` be the immersion
        :math:`f: M \rightarrow \mathbb{R}^n` of the manifold
        :math:`M` into the Euclidean space :math:`\mathbb{R}^n`.
        The elements of the metric matrix at a base point :math:`p`
        are defined as:
        :math:`(f*g)_{ij}(p) = <df_p e_i , df_p e_j>`,
        for :math:`e_i, e_j` basis elements of :math:`M`.

        Parameters
        ----------
        base_point : array-like, shape=[..., dim]
            Base point.
            Optional, default: None.

        Returns
        -------
        mat : array-like, shape=[..., dim, dim]
            Inner-product matrix.
        """
        dim, dim_embedding = self._space.dim, self._space.embedding_space.dim
        is_vec = check_is_batch(self._space.point_ndim, base_point)

        immersed_base_point = self._space.immersion(base_point)
        jacobian_immersion = self._space.jacobian_immersion(base_point)

        basis_elements = gs.eye(dim)

        if is_vec:
            reshaped_jacobian_immersion = gs.reshape(jacobian_immersion, (-1, dim))
            reshaped_immersed_basis_elements = gs.matvec(
                reshaped_jacobian_immersion, basis_elements
            )
            immersed_basis_elements = gs.moveaxis(
                gs.reshape(reshaped_immersed_basis_elements, (dim, -1, dim_embedding)),
                0,
                1,
            )

        else:
            immersed_basis_elements = gs.matvec(jacobian_immersion, basis_elements)

        elems = {}
        for i in range(dim):
            for j in range(i, dim):
                elem = self._space.embedding_space.metric.inner_product(
                    immersed_basis_elements[..., i, :],
                    immersed_basis_elements[..., j, :],
                    immersed_base_point,
                )
                elems[(i, j)] = elem

        mat = []
        for i in range(dim):
            for j in range(dim):
                elem = elems[(i, j)] if j > i else elems[(j, i)]
                mat.append(elem)

        shape = (-1, dim, dim) if is_vec else (dim, dim)
        return gs.reshape(gs.stack(mat, axis=-1), shape)

    def inner_product_derivative_matrix(self, base_point):
        r"""Compute the inner-product derivative matrix.

        The derivative of the metrix matrix is given by
        :math:`\partial_k g_{ij}(p)`
        where :math:`p` is the base_point.

        The index k of the derivation is last.

        Parameters
        ----------
        base_point : array-like, shape=[..., *shape]
            Base point.
            Optional, default: None.

        Returns
        -------
        inner_prod_deriv_mat : array-like, shape=[..., dim, dim, dim]
            Inner-product derivative matrix, where the index of the derivation
            is last: :math:`mat_{ijk} = \partial_k g_{ij}`.
        """
        jacobian_ai = self._space.jacobian_immersion(base_point)
        hessian_aij = self._space.hessian_immersion(base_point)
        return gs.einsum("...aki,...aj->...ijk", hessian_aij, jacobian_ai) + gs.einsum(
            "...akj,...ai->...ijk", hessian_aij, jacobian_ai
        )

    def second_fundamental_form(self, base_point):
        r"""Compute the second fundamental form.

        In the case of an immersion f, the second fundamental form is
        given by the formula:
        :math:`\RN{2}(p)_{ij}^\alpha = \partial_{i j}^2 f^\alpha(p)`
        :math:`-\Gamma_{i j}^k(p) \partial_k f^\alpha(p)`
        at base_point :math:`p`.

        Parameters
        ----------
        base_point : array-like, shape=[..., dim]
            Base point.

        Returns
        -------
        second_fundamental_form : array-like, shape=[..., embedding_dim, dim, dim]
            Second fundamental form :math:`\RN{2}(p)_{ij}^\alpha` where the
            :math:`\alpha` index is first.
        """
        christoffels = self.christoffels(base_point)

        jacobian = self._space.jacobian_immersion(base_point)
        hessian = self._space.hessian_immersion(base_point)

        return hessian - gs.einsum("...kij,...dk->...dij", christoffels, jacobian)

    def mean_curvature_vector(self, base_point):
        r"""Compute the mean curvature vector.

        The mean curvature vector is defined at base point :math:`p` by
        :math:`H_p^\alpha= \frac{1}{d} (f^{*}g)_{p}^{ij} (\partial_{i j}^2 f^\alpha(p)`
        :math:`-\Gamma_{i j}^k(p) \partial_k f^\alpha(p))`
        where :math:`f^{*}g` is the pullback of the metric :math:`g` by the
        immersion :math:`f`.

        Parameters
        ----------
        base_point : array-like, shape=[..., dim]
            Base point.

        Returns
        -------
        mean_curvature_vector : array-like, shape=[..., embedding_dim]
            Mean curvature vector.
        """
        second_fund_form = self.second_fundamental_form(base_point)
        cometric = self.cometric_matrix(base_point)
        return gs.einsum("...ij,...aij->...a", cometric, second_fund_form)


class PullbackDiffeoMetric(RiemannianMetric):
    """Pullback metric induced by a diffeomorphism."""

    def __init__(self, space, diffeo, image_space, signature=None):
        super().__init__(space, signature=signature)
        self.diffeo = diffeo
        self.image_space = image_space

    def metric_matrix(self, base_point=None):
        """Metric matrix at the tangent space at a base point.

        Parameters
        ----------
        base_point : array-like, shape=[..., *shape]
            Base point.

        Returns
        -------
        mat : array-like, shape=[..., dim, dim]
            Inner-product matrix.
        """
        raise NotImplementedError(
            "The computation of the pullback metric matrix"
            " is not implemented yet in general shape setting."
        )

    def inner_product(self, tangent_vec_a, tangent_vec_b, base_point=None):
        """Inner product between two tangent vectors at a base point.

        Parameters
        ----------
        tangent_vec_a: array-like, shape=[..., *shape]
            Tangent vector at base point.
        tangent_vec_b: array-like, shape=[..., *shape]
            Tangent vector at base point.
        base_point: array-like, shape=[..., *shape]
            Base point.

        Returns
        -------
        inner_product : array-like, shape=[...,]
            Inner-product.
        """
        image_point = self.diffeo(base_point)
        return self.image_space.metric.inner_product(
            self.diffeo.tangent(
                tangent_vec_a,
                base_point=base_point,
                image_point=image_point,
            ),
            self.diffeo.tangent(
                tangent_vec_b,
                base_point=base_point,
                image_point=image_point,
            ),
            image_point,
        )

    def squared_norm(self, vector, base_point=None):
        """Compute the square of the norm of a vector.

        Squared norm of a vector associated to the inner product
        at the tangent space at a base point.

        Parameters
        ----------
        vector : array-like, shape=[..., dim]
            Vector.
        base_point : array-like, shape=[..., dim]
            Base point.

        Returns
        -------
        sq_norm : array-like, shape=[...,]
            Squared norm.
        """
        image_point = self.diffeo(base_point)
        return self.image_space.metric.squared_norm(
            self.diffeo.tangent(
                vector,
                base_point=base_point,
                image_point=image_point,
            ),
            image_point,
        )

    def norm(self, vector, base_point=None):
        """Compute norm of a vector.

        Norm of a vector associated to the inner product
        at the tangent space at a base point.

        Note: This only works for positive-definite
        Riemannian metrics and inner products.

        Parameters
        ----------
        vector : array-like, shape=[..., dim]
            Vector.
        base_point : array-like, shape=[..., dim]
            Base point.

        Returns
        -------
        norm : array-like, shape=[...,]
            Norm.
        """
        image_point = self.diffeo(base_point)
        return self.image_space.metric.norm(
            self.diffeo.tangent(
                vector,
                base_point=base_point,
                image_point=image_point,
            ),
            image_point,
        )

    def exp(self, tangent_vec, base_point):
        """Compute the exponential map via diffeomorphic pullback.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., *shape]
            Tangent vector.
        base_point : array-like, shape=[..., *shape]
            Base point.

        Returns
        -------
        exp : array-like, shape=[..., *shape]
            Riemannian exponential of tangent_vec at base_point.
        """
        image_base_point = self.diffeo(base_point)
        image_tangent_vec = self.diffeo.tangent(
            tangent_vec, base_point=base_point, image_point=image_base_point
        )
        image_exp = self.image_space.metric.exp(
            image_tangent_vec,
            image_base_point,
        )
        return self.diffeo.inverse(image_exp)

    def log(self, point, base_point):
        """Compute the logarithm map via diffeomorphic pullback.

        Parameters
        ----------
        point : array-like, shape=[..., *shape]
            Point.
        base_point : array-like, shape=[..., *shape]
            Point.

        Returns
        -------
        log : array-like, shape=[..., *shape]
            Logarithm of point from base_point.
        """
        image_base_point = self.diffeo(base_point)
        image_point = self.diffeo(point)
        image_log = self.image_space.metric.log(image_point, image_base_point)
        return self.diffeo.inverse_tangent(
            image_log, image_point=image_base_point, base_point=base_point
        )

    def squared_dist(self, point_a, point_b):
        """Squared geodesic distance between two points via pullback.

        Parameters
        ----------
        point_a : array-like, shape=[..., dim]
            Point.
        point_b : array-like, shape=[..., dim]
            Point.

        Returns
        -------
        sq_dist : array-like, shape=[...,]
            Squared distance.
        """
        image_point_a = self.diffeo(point_a)
        image_point_b = self.diffeo(point_b)
        return self.image_space.metric.squared_dist(image_point_a, image_point_b)

    def dist(self, point_a, point_b):
        """Compute the distance via diffeomorphic pullback.

        Parameters
        ----------
        point_a : array-like, shape=[..., *shape]
            Point a.
        point_b : array-like, shape=[..., *shape]
            Point b.

        Returns
        -------
        dist : array-like
            Distance between point_a and point_b.
        """
        image_point_a = self.diffeo(point_a)
        image_point_b = self.diffeo(point_b)
        return self.image_space.metric.dist(image_point_a, image_point_b)

    def geodesic(self, initial_point, end_point=None, initial_tangent_vec=None):
        """Compute the geodesic via diffeomorphic pullback.

        Parameters
        ----------
        initial_point : array-like, shape=[..., *shape]
            Initial Point.
        end_point : array-like, shape=[..., *shape]
            End Point.

        Returns
        -------
        geodesic : callable
            Geodesic between initial_point and end_point.
        """
        image_initial_point = self.diffeo(initial_point)
        if end_point is None:
            image_end_point = None
        else:
            image_end_point = self.diffeo(end_point)
        if initial_tangent_vec is None:
            image_initial_tangent_vec = None
        else:
            image_initial_tangent_vec = self.diffeo.tangent(
                initial_tangent_vec,
                base_point=initial_point,
                image_point=image_initial_point,
            )
        image_geodesic = self.image_space.metric.geodesic(
            image_initial_point, image_end_point, image_initial_tangent_vec
        )

        def path(t):
            image_point_t = image_geodesic(t)
            point_t = self.diffeo.inverse(image_point_t)
            return point_t

        return path

    def curvature(self, tangent_vec_a, tangent_vec_b, tangent_vec_c, base_point=None):
        """Compute the curvature via diffeomorphic pullback.

        Parameters
        ----------
        tangent_vec_a : array-like, shape=[..., *shape]
            Tangent vector a.
        tangent_vec_b : array-like, shape=[..., *shape]
            Tangent vector b.
        tangent_vec_c : array-like, shape=[..., *shape]
            Tangent vector c.
        base_point : array-like, shape=[..., *shape]
            Base point.

        Returns
        -------
        curvature : array-like, shape=[..., *shape]
            Curvature in directions tangent_vec a, b, c at base_point.
        """
        image_base_point = self.diffeo(base_point)
        image_tangent_vec_a = self.diffeo.tangent(
            tangent_vec_a,
            base_point=base_point,
            image_point=image_base_point,
        )
        image_tangent_vec_b = self.diffeo.tangent(
            tangent_vec_b,
            base_point=base_point,
            image_point=image_base_point,
        )
        image_tangent_vec_c = self.diffeo.tangent(
            tangent_vec_c,
            base_point=base_point,
            image_point=image_base_point,
        )
        image_curvature = self.image_space.metric.curvature(
            image_tangent_vec_a,
            image_tangent_vec_b,
            image_tangent_vec_c,
            image_base_point,
        )
        return self.diffeo.inverse_tangent(
            image_curvature, image_point=image_base_point, base_point=base_point
        )

    def parallel_transport(
        self, tangent_vec, base_point, direction=None, end_point=None
    ):
        """Compute the parallel transport via diffeomorphic pullback.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., *shape]
            Tangent vector.
        base_point : array-like, shape=[..., *shape]
            Base point.
        direction : array-like, shape=[..., *shape]
            Direction.
        end_point : array-like, shape=[..., *shape]
            End point.

        Returns
        -------
        parallel_transport : array-like, shape=[..., *shape]
            Parallel transport.
        """
        image_base_point = self.diffeo(base_point)
        image_tangent_vec = self.diffeo.tangent(
            tangent_vec, base_point=base_point, image_point=image_base_point
        )

        image_direction = (
            None
            if direction is None
            else self.diffeo.tangent(
                direction, base_point=base_point, image_point=image_base_point
            )
        )
        image_end_point = None if end_point is None else self.diffeo(end_point)

        image_parallel_transport = self.image_space.metric.parallel_transport(
            image_tangent_vec,
            image_base_point,
            direction=image_direction,
            end_point=image_end_point,
        )

        if image_end_point is None:
            image_end_point = self.image_space.metric.exp(
                image_direction, image_base_point
            )

        return self.diffeo.inverse_tangent(
            image_parallel_transport, image_point=image_end_point, base_point=end_point
        )



"""Class to implement simply the Fisher-Rao metric on information manifolds."""

from scipy.integrate import quad_vec

import geomstats.backend as gs
from geomstats.geometry.riemannian_metric import RiemannianMetric


class FisherRaoMetric(RiemannianMetric):
    r"""Class to derive the information metric from the pdf in InformationManifoldMixin.

    Given a statistical manifold with coordinates :math:`\theta`,
    the Fisher information metric is:

    .. math::
        g_{j k}(\theta)=\int_X \frac{\partial \log p(x, \theta)}
        {\partial \theta_j}\frac{\partial \log p(x, \theta)}
        {\partial \theta_k} p(x, \theta) d x

    Attributes
    ----------
    space : InformationManifold
        Riemannian Manifold for a parametric family of (real) distributions.
    support : list, shape = (2,)
        Left and right bounds for the support of the distribution.
        But this is just to help integration, bounds should be as large as needed.
    """

    def __init__(self, space, support):
        super().__init__(
            space=space,
            signature=(space.dim, 0),
        )
        self.support = support

    def metric_matrix(self, base_point):
        r"""Compute the inner-product matrix.

        The Fisher information matrix is noted I in the literature.

        Compute the inner-product matrix of the Fisher information metric
        at the tangent space at base point.

        .. math::
            I(\theta) = \mathbb{E}[YY^T]

        where,

        .. math::
            Y = \nabla_{\theta} \log f_{\theta}(x)

        After manipulation and in indicial notation

        .. math::
             I_{ij} = \int \
             \partial_{i} f_{\theta}(x)\
             \partial_{j} f_{\theta}(x)\
             \frac{1}{f_{\theta}(x)}

        Parameters
        ----------
        base_point : array-like, shape=[..., dim]
            Base point.

        Returns
        -------
        metric_mat : array-like, shape=[..., dim, dim]
            Inner-product matrix.

        References
        ----------
        .. [AS1985] Amari, S (1985)
            Differential Geometric Methods in Statistics, Berlin, Springer – Verlag.
        """

        def pdf(x):
            """Compute pdf at a fixed point on the support.

            Parameters
            ----------
            x : float, shape (,)
                Point on the support of the distribution
            """
            return lambda point: gs.squeeze(self._space.point_to_pdf(point)(x), axis=-1)

        def _function_to_integrate(x):
            pdf_x = pdf(x)
            (
                pdf_x_at_base_point,
                pdf_x_derivative_at_base_point,
            ) = gs.autodiff.value_and_grad(pdf_x)(base_point)

            return gs.einsum(
                "...ij,...->...ij",
                gs.einsum(
                    "...i,...j->...ij",
                    pdf_x_derivative_at_base_point,
                    pdf_x_derivative_at_base_point,
                ),
                1 / pdf_x_at_base_point,
            )

        return quad_vec(_function_to_integrate, *self.support)[0]

    def inner_product_derivative_matrix(self, base_point):
        r"""Compute the derivative of the inner-product matrix.

        Compute the derivative of the inner-product matrix of
        the Fisher information metric at the tangent space at base point.

        .. math::
            \partial_{\theta} I(\theta) = \partial_{\theta} \mathbb{E}[YY^T]

        where,

        .. math::
            Y = \nabla_{\theta} \log f_{\theta}(x)

        or, in indicial notation:

        .. math::
            \partial_k I_{ij} = \int\
            \partial_{ki}^2 f\partial_j f \frac{1}{f} + \
            \partial_{kj}^2 f\partial_i f \frac{1}{f} - \
            \partial_i f \partial_j f \partial_k f \frac{1}{f^2}

        with :math:`f = f_{\theta}(x)`

        Parameters
        ----------
        base_point : array-like, shape=[..., dim]
            Base point.

        Returns
        -------
        mat : array-like, shape=[..., dim, dim, dim]
            Derivative of the inner-product matrix, where the index
            k of the derivation is last: math:`mat_{ijk} = \partial_k g_{ij}`.
        """

        def pdf(x):
            """Compute pdf at a fixed point on the support.

            Parameters
            ----------
            x : float, shape (,)
                Point on the support of the distribution
            """
            return lambda point: gs.squeeze(self._space.point_to_pdf(point)(x), axis=-1)

        def _function_to_integrate(x):
            pdf_x = pdf(x)
            (
                pdf_x_at_base_point,
                pdf_x_derivative_at_base_point,
                pdf_x_hessian_at_base_point,
            ) = gs.autodiff.value_jacobian_and_hessian(pdf_x)(base_point)

            return gs.einsum(
                "...,...ijk->...ijk",
                1 / (pdf_x_at_base_point**2),
                gs.einsum(
                    "...,...ijk->...ijk",
                    pdf_x_at_base_point,
                    gs.einsum(
                        "...ki,...j->...ijk",
                        pdf_x_hessian_at_base_point,
                        pdf_x_derivative_at_base_point,
                    )
                    + gs.einsum(
                        "...kj,...i->...ijk",
                        pdf_x_hessian_at_base_point,
                        pdf_x_derivative_at_base_point,
                    ),
                )
                - gs.einsum(
                    "...i, ...j, ...k -> ...ijk",
                    pdf_x_derivative_at_base_point,
                    pdf_x_derivative_at_base_point,
                    pdf_x_derivative_at_base_point,
                ),
            )

        return quad_vec(_function_to_integrate, *self.support)[0]



"""The MDM classifier on manifolds.

Lead authors: Daniel Brooks and Quentin Barthelemy.
"""

from scipy.special import softmax
from sklearn.base import BaseEstimator, ClassifierMixin

import geomstats.backend as gs
from geomstats.learning._template import TransformerMixin
from geomstats.learning.frechet_mean import FrechetMean


class RiemannianMinimumDistanceToMean(
    BaseEstimator,
    ClassifierMixin,
    TransformerMixin,
):
    """Minimum Distance to Mean (MDM) classifier on manifolds.

    Classification by nearest centroid. For each of the given classes, a
    centroid is estimated according to the chosen metric. Then, for each new
    point, the class is affected according to the nearest centroid [BBCJ2012]_.

    Parameters
    ----------
    space : Manifold
        Equipped manifold.

    Attributes
    ----------
    classes_ : array-like, shape=[n_classes,]
        If fit, labels of training set.
    mean_estimates_ : array-like, shape=[n_classes, *space.shape]
        If fit, centroids computed on training set.

    Notes
    -----
    * Required metric methods: `squared_dist`, `closest_neighbot_index`.

    References
    ----------
    .. [BBCJ2012] A. Barachant, S. Bonnet, M. Congedo and C. Jutten, Multiclass
        Brain-Computer Interface Classification by Riemannian Geometry. IEEE
        Trans. Biomed. Eng., vol. 59, pp. 920-928, 2012.
    """

    def __init__(self, space):
        self.space = space

        self.classes_ = None
        self.mean_estimates_ = None

        self.mean_estimator = FrechetMean(space)

    @property
    def n_classes_(self):
        """Number of classes."""
        return len(self.classes_)

    def fit(self, X, y, weights=None):
        """Compute Frechet mean of each class.

        Parameters
        ----------
        X : array-like, shape=[n_samples, *space.shape]
            Training input samples.
        y : array-like, shape=[n_samples,]
            Training labels.
        weights : array-like, shape=[n_samples,]
            Weights associated to the samples.
            Optional, default: None, in which case it is equally weighted.

        Returns
        -------
        self : object
            Returns self.
        """
        self.classes_ = gs.unique(y)
        if weights is None:
            weights = gs.ones(X.shape[0])
        weights /= gs.sum(weights)

        frechet_means = []
        for c in self.classes_:
            X_c = X[gs.where(y == c, True, False)]
            weights_c = weights[gs.where(y == c, True, False)]
            mean_c = self.mean_estimator.fit(X_c, weights=weights_c).estimate_
            frechet_means.append(mean_c)

        self.mean_estimates_ = gs.array(frechet_means)

        return self

    def predict(self, X):
        """Compute closest neighbor according to riemannian_metric.

        Parameters
        ----------
        X : array-like, shape=[n_samples, *space.shape]
            Test samples.

        Returns
        -------
        y : array-like, shape=[n_samples,]
            Predicted labels.
        """
        indices = self.space.metric.closest_neighbor_index(
            X,
            self.mean_estimates_,
        )
        if gs.ndim(indices) == 0:
            indices = gs.expand_dims(indices, 0)

        return gs.to_ndarray(gs.take(self.classes_, indices), to_ndim=1)

    def predict_proba(self, X):
        """Compute probabilities.

        Compute probabilities to belong to classes according to
        riemannian_metric.

        Parameters
        ----------
        X : array-like, shape=[n_samples, *space.shape]
            Test samples.

        Returns
        -------
        probas : array-like, shape=[n_samples, n_classes]
            Probability of the sample for each class in the model.
        """
        dists2 = gs.stack(
            [
                self.space.metric.squared_dist(
                    X,
                    mean_estimate,
                )
                for mean_estimate in self.mean_estimates_
            ],
            axis=-1,
        )
        probas = softmax(-dists2, axis=-1)
        return gs.from_numpy(probas)

    def transform(self, X):
        """Compute distances to each centroid.

        Compute distances to each centroid according to riemannian_metric.

        Parameters
        ----------
        X : array-like, shape=[n_samples, *space.shape]
            Test samples.

        Returns
        -------
        dist : ndarray, shape=[n_samples, n_classes]
            Distances to each centroid.
        """
        dists = gs.stack(
            [
                self.space.metric.dist(
                    X,
                    mean_estimate,
                )
                for mean_estimate in self.mean_estimates_
            ],
            axis=-1,
        )
        return dists



"""Manifold for sets of landmarks that belong to any given manifold.

Lead author: Nicolas Guigui.
"""

from geomstats.geometry.nfold_manifold import NFoldManifold, NFoldMetric


class Landmarks(NFoldManifold):
    """Class for space of landmarks.

    The landmark space is a product manifold where all manifolds in the
    product are the same. The default metric is the product metric and
    is often referred to as the L2 metric.

    Parameters
    ----------
    ambient_manifold : Manifold
        Manifold to which landmarks belong.
    k_landmarks : int
        Number of landmarks.
    """

    def __init__(self, ambient_manifold, k_landmarks, equip=True):
        super().__init__(
            base_manifold=ambient_manifold, n_copies=k_landmarks, equip=equip
        )

    @property
    def k_landmarks(self):
        """Number of landmarks."""
        return self.n_copies

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return L2LandmarksMetric


class L2LandmarksMetric(NFoldMetric):
    """L2 Riemannian metric on the space of landmarks.

    This is the NFoldMetric of the n-fold manifold made out
    of k_landmarks copies of the ambient manifold of each landmark.
    """



"""N-fold product manifold.

Lead author: Nicolas Guigui, John Harvey.
"""

import geomstats.backend as gs
import geomstats.errors
from geomstats.geometry.manifold import Manifold
from geomstats.geometry.riemannian_metric import RiemannianMetric
from geomstats.vectorization import get_batch_shape


class NFoldManifold(Manifold):
    r"""Class for an n-fold product manifold :math:`M^n`.

    Define a manifold as the product manifold of n copies of a given base
    manifold M.

    Parameters
    ----------
    base_manifold : Manifold
        Base manifold.
    n_copies : int
        Number of replication of the base manifold.
    """

    def __init__(
        self,
        base_manifold,
        n_copies,
        equip=True,
    ):
        geomstats.errors.check_integer(n_copies, "n_copies")
        dim = n_copies * base_manifold.dim
        shape = (n_copies,) + base_manifold.shape

        self.base_manifold = base_manifold
        self.n_copies = n_copies

        super().__init__(
            dim=dim,
            shape=shape,
            intrinsic=base_manifold.intrinsic,
            equip=equip,
        )

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return NFoldMetric

    def belongs(self, point, atol=gs.atol):
        """Test if a point belongs to the manifold.

        Parameters
        ----------
        point : array-like, shape=[..., n_copies, *base_shape]
            Point.
        atol : float,
            Tolerance.

        Returns
        -------
        belongs : array-like, shape=[..., n_copies, *base_shape]
            Boolean evaluating if the point belongs to the manifold.
        """
        batch_shape = get_batch_shape(self.point_ndim, point)
        point_ = gs.reshape(point, (-1, *self.base_manifold.shape))

        each_belongs = self.base_manifold.belongs(point_, atol=atol)

        reshaped = gs.reshape(each_belongs, batch_shape + (self.n_copies,))
        return gs.all(reshaped, axis=-1)

    def is_tangent(self, vector, base_point, atol=gs.atol):
        """Check whether the vector is tangent at base_point.

        The tangent space of the product manifold is the direct sum of
        tangent spaces.

        Parameters
        ----------
        vector : array-like, shape=[..., n_copies, *base_shape]
            Vector.
        base_point : array-like, shape=[..., n_copies, *base_shape]
            Point on the manifold.
        atol : float
            Absolute tolerance.
            Optional, default: backend atol.

        Returns
        -------
        is_tangent : bool
            Boolean denoting if vector is a tangent vector at the base point.
        """
        vector_, base_point_ = gs.broadcast_arrays(vector, base_point)
        batch_shape = get_batch_shape(self.point_ndim, vector_)
        base_point_ = gs.reshape(base_point_, (-1, *self.base_manifold.shape))
        vector_ = gs.reshape(vector_, (-1, *self.base_manifold.shape))

        each_tangent = self.base_manifold.is_tangent(vector_, base_point_, atol=atol)

        reshaped = gs.reshape(each_tangent, batch_shape + (self.n_copies,))
        return gs.all(reshaped, axis=-1)

    def to_tangent(self, vector, base_point):
        """Project a vector to a tangent space of the manifold.

        The tangent space of the product manifold is the direct sum of
        tangent spaces.

        Parameters
        ----------
        vector : array-like, shape=[..., n_copies, *base_shape]
            Vector.
        base_point : array-like, shape=[..., n_copies, *base_shape]
            Point on the manifold.

        Returns
        -------
        tangent_vec : array-like, shape=[..., n_copies, *base_shape]
            Tangent vector at base point.
        """
        vector_, base_point_ = gs.broadcast_arrays(vector, base_point)
        base_point_ = gs.reshape(base_point_, (-1, *self.base_manifold.shape))
        batch_shape = get_batch_shape(self.point_ndim, vector_)
        vector_ = gs.reshape(vector_, (-1, *self.base_manifold.shape))

        each_tangent = self.base_manifold.to_tangent(vector_, base_point_)

        return gs.reshape(each_tangent, batch_shape + self.shape)

    def random_point(self, n_samples=1, bound=1.0):
        """Sample in the product space from the product distribution.

        The distribution used is the product of the distributions that each copy of the
        manifold uses in its own random_point method.

        Parameters
        ----------
        n_samples : int, optional
            Number of samples.
        bound : float
            Bound of the interval in which to sample for non compact manifolds.
            Optional, default: 1.

        Returns
        -------
        samples : array-like, shape=[..., n_copies, *base_shape]
            Points sampled on the product manifold.
        """
        sample = self.base_manifold.random_point(n_samples * self.n_copies, bound)
        reshaped = gs.reshape(
            sample, (n_samples, self.n_copies) + self.base_manifold.shape
        )
        if n_samples > 1:
            return reshaped
        return gs.squeeze(reshaped, axis=0)

    def projection(self, point):
        """Project a point from product embedding manifold to the product manifold.

        Parameters
        ----------
        point : array-like, shape=[..., n_copies, *base_shape]
            Point in embedding manifold.

        Returns
        -------
        projected : array-like, shape=[..., n_copies, *base_shape]
            Projected point.
        """
        if hasattr(self.base_manifold, "projection"):
            batch_shape = get_batch_shape(self.point_ndim, point)
            point_ = gs.reshape(point, (-1, *self.base_manifold.shape))
            projected = self.base_manifold.projection(point_)
            return gs.reshape(projected, batch_shape + self.shape)
        raise NotImplementedError(
            "The base manifold does not implement a projection method."
        )


class NFoldMetric(RiemannianMetric):
    r"""Class for an n-fold product manifold :math:`M^n`.

    Define a manifold as the product manifold of n copies of a given base
    manifold M.

    Parameters
    ----------
    space : NFoldManifold
        Base space.
    scales : array-like
        Scale of each metric in the product.
    """

    def __init__(self, space, scales=None, signature=None):
        if scales is not None:
            for scale in scales:
                geomstats.errors.check_positive(scale, "Each value in scales")

            if len(scales) != space.n_copies:
                raise ValueError(
                    "Number of scales should be equal to number of factors"
                )
        self.scales = scales

        super().__init__(space=space, signature=signature)

    def metric_matrix(self, base_point):
        """Compute the matrix of the inner-product.

        Matrix of the inner-product defined by the Riemmanian metric
        at point base_point of the manifold.

        Parameters
        ----------
        base_point : array-like, shape=[..., n_copies, *base_shape]
            Point on the manifold at which to compute the inner-product matrix.

        Returns
        -------
        matrix : array-like, shape=[..., n_copies, dim, dim]
            Matrix of the inner-product at the base point.
        """
        base_manifold = self._space.base_manifold
        batch_shape = get_batch_shape(self._space.point_ndim, base_point)

        point_ = gs.reshape(base_point, (-1, *base_manifold.shape))
        matrices = base_manifold.metric.metric_matrix(point_)

        dim = base_manifold.shape[-1]
        reshaped = gs.reshape(matrices, batch_shape + (self._space.n_copies, dim, dim))

        if self.scales is not None:
            reshaped = gs.einsum("j,...jkl->...jkl", self.scales, reshaped)

        return reshaped

    def pointwise_inner_product(self, tangent_vec_a, tangent_vec_b, base_point):
        """Pointwise inner product.

        Parameters
        ----------
        tangent_vec_a : array-like, shape=[..., n_copies, *base_shape]
            First tangent vector at base point.
        tangent_vec_b : array-like, shape=[..., n_copies, *base_shape]
            Second tangent vector at base point.
        base_point : array-like, shape=[..., n_copies, *base_shape]
            Point on the manifold.
            Optional, default: None.

        Returns
        -------
        pointwise_inner_prod : array-like, shape=[..., n_copies]
            Inner-product of the two tangent vectors.
        """
        base_manifold = self._space.base_manifold

        tangent_vec_a_, tangent_vec_b_, point_ = gs.broadcast_arrays(
            tangent_vec_a, tangent_vec_b, base_point
        )
        batch_shape = get_batch_shape(self._space.point_ndim, tangent_vec_a_)

        point_ = gs.reshape(point_, (-1, *base_manifold.shape))
        vector_a = gs.reshape(tangent_vec_a_, (-1, *base_manifold.shape))
        vector_b = gs.reshape(tangent_vec_b_, (-1, *base_manifold.shape))
        inner_each = base_manifold.metric.inner_product(vector_a, vector_b, point_)

        reshaped = gs.reshape(inner_each, batch_shape + (self._space.n_copies,))

        if self.scales is not None:
            reshaped = gs.einsum("j,...j->...j", self.scales, reshaped)

        return reshaped

    def inner_product(self, tangent_vec_a, tangent_vec_b, base_point):
        """Compute the inner-product of two tangent vectors at a base point.

        Inner product defined by the Riemannian metric at point `base_point`
        between tangent vectors `tangent_vec_a` and `tangent_vec_b`.

        Parameters
        ----------
        tangent_vec_a : array-like, shape=[..., n_copies, *base_shape]
            First tangent vector at base point.
        tangent_vec_b : array-like, shape=[..., n_copies, *base_shape]
            Second tangent vector at base point.
        base_point : array-like, shape=[..., n_copies, *base_shape]
            Point on the manifold.
            Optional, default: None.

        Returns
        -------
        inner_prod : array-like, shape=[...,]
            Inner-product of the two tangent vectors.
        """
        return gs.sum(
            self.pointwise_inner_product(tangent_vec_a, tangent_vec_b, base_point),
            axis=-1,
        )

    def pointwise_norm(self, tangent_vec, base_point):
        """Compute the pointwise norms of a tangent vector.

        Compute the norms of the components of a tangent vector at the different
        sampling points of a base curve.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., n_copies, *base_shape]
            Tangent vector to discrete curve.
        base_point : array-like, shape=[..., n_copies, *base_shape]
            Point representing a discrete curve.

        Returns
        -------
        norm : array-like, shape=[..., *base_shape]
            Point-wise norms.
        """
        sq_norm = self.pointwise_inner_product(
            tangent_vec_a=tangent_vec, tangent_vec_b=tangent_vec, base_point=base_point
        )
        return gs.sqrt(sq_norm)

    def exp(self, tangent_vec, base_point):
        """Compute the Riemannian exponential of a tangent vector.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., n_copies, *base_shape]
            Tangent vector at a base point.
        base_point : array-like, shape=[..., n_copies, *base_shape]
            Point on the manifold.
            Optional, default: None.

        Returns
        -------
        exp : array-like, shape=[..., n_copies, *base_shape]
            Point on the manifold equal to the Riemannian exponential
            of tangent_vec at the base point.
        """
        base_manifold = self._space.base_manifold
        batch_shape = get_batch_shape(self._space.point_ndim, tangent_vec, base_point)

        tangent_vec, point_ = gs.broadcast_arrays(tangent_vec, base_point)
        point_ = gs.reshape(point_, (-1, *base_manifold.shape))
        vector_ = gs.reshape(tangent_vec, (-1, *base_manifold.shape))
        each_exp = base_manifold.metric.exp(vector_, point_)
        return gs.reshape(each_exp, batch_shape + self._space.shape)

    def log(self, point, base_point):
        """Compute the Riemannian logarithm of a point.

        Parameters
        ----------
        point : array-like, shape=[..., n_copies, *base_shape]
            Point on the manifold.
        base_point : array-like, shape=[..., n_copies, *base_shape]
            Point on the manifold.
            Optional, default: None.

        Returns
        -------
        log : array-like, shape=[..., n_copies, *base_shape]
            Tangent vector at the base point equal to the Riemannian logarithm
            of point at the base point.
        """
        base_manifold = self._space.base_manifold
        batch_shape = get_batch_shape(self._space.point_ndim, point, base_point)

        point_, base_point_ = gs.broadcast_arrays(point, base_point)
        base_point_ = gs.reshape(base_point_, (-1, *base_manifold.shape))
        point_ = gs.reshape(point_, (-1, *base_manifold.shape))
        each_log = base_manifold.metric.log(point_, base_point_)
        return gs.reshape(each_log, batch_shape + self._space.shape)
